# -*- coding: utf-8 -*-
import logging
import collections

import six

import idb.netnode
import idb.analysis


logger = logging.getLogger(__name__)


def is_flag_set(flags, flag):
    return flags & flag == flag


class FLAGS:
    # instruction/data operands
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__op.html

    # outer offset base (combined with operand number). More...
    OPND_OUTER = 0x80

    # mask for operand number
    OPND_MASK = 0x07

    # all operands
    OPND_ALL = OPND_MASK

    # byte states bits
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__statebits.html

    # Mask for typing.
    MS_CLS = 0x00000600

    # Code ?
    FF_CODE = 0x00000600

    # Data ?
    FF_DATA = 0x00000400

    # Tail ?
    FF_TAIL = 0x00000200

    # Unknown ?
    FF_UNK = 0x00000000

    # specific state information bits
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__statespecb.html

    # Mask of common bits.
    MS_COMM = 0x000FF800

    # Has comment ?
    FF_COMM = 0x00000800

    # has references
    FF_REF = 0x00001000

    # Has next or prev lines ?
    FF_LINE = 0x00002000

    # Has name ?
    FF_NAME = 0x00004000

    # Has dummy name?
    FF_LABL = 0x00008000

    # Exec flow from prev instruction.
    FF_FLOW = 0x00010000

    # Inverted sign of operands.
    FF_SIGN = 0x00020000

    # Bitwise negation of operands.
    FF_BNOT = 0x00040000

    # is variable byte?
    FF_VAR = 0x00080000

    # instruction operand types bites
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__opbits.html

    # Mask for 1st arg typing.
    MS_0TYPE = 0x00F00000

    # Void (unknown)?
    FF_0VOID = 0x00000000

    # Hexadecimal number?
    FF_0NUMH = 0x00100000

    # Decimal number?
    FF_0NUMD = 0x00200000

    # Char ('x')?
    FF_0CHAR = 0x00300000

    # Segment?
    FF_0SEG = 0x00400000

    # Offset?
    FF_0OFF = 0x00500000

    # Binary number?
    FF_0NUMB = 0x00600000

    # Octal number?
    FF_0NUMO = 0x00700000

    # Enumeration?
    FF_0ENUM = 0x00800000

    # Forced operand?
    FF_0FOP = 0x00900000

    # Struct offset?
    FF_0STRO = 0x00A00000

    # Stack variable?
    FF_0STK = 0x00B00000

    # Floating point number?
    FF_0FLT = 0x00C00000

    # Custom representation?
    FF_0CUST = 0x00D00000

    # Mask for the type of other operands.
    MS_1TYPE = 0x0F000000

    # Void (unknown)?
    FF_1VOID = 0x00000000

    # Hexadecimal number?
    FF_1NUMH = 0x01000000

    # Decimal number?
    FF_1NUMD = 0x02000000

    # Char ('x')?
    FF_1CHAR = 0x03000000

    # Segment?
    FF_1SEG = 0x04000000

    # Offset?
    FF_1OFF = 0x05000000

    # Binary number?
    FF_1NUMB = 0x06000000

    # Octal number?
    FF_1NUMO = 0x07000000

    # Enumeration?
    FF_1ENUM = 0x08000000

    # Forced operand?
    FF_1FOP = 0x09000000

    # Struct offset?
    FF_1STRO = 0x0A000000

    # Stack variable?
    FF_1STK = 0x0B000000

    # Floating point number?
    FF_1FLT = 0x0C000000

    # Custom representation?
    FF_1CUST = 0x0D000000

    # code byte bits
    # via: https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__codebits.html
    # Mask for code bits.
    MS_CODE = 0xF0000000

    # function start?
    FF_FUNC = 0x10000000

    # Has Immediate value?
    FF_IMMD = 0x40000000

    # Has jump table or switch_info?
    FF_JUMP = 0x80000000

    # data bytes bits
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__databits.html

    # Mask for DATA typing.
    DT_TYPE = 0xF0000000

    # byte
    FF_BYTE = 0x00000000

    # word
    FF_WORD = 0x10000000

    # double word
    FF_DWRD = 0x20000000

    # quadro word
    FF_QWRD = 0x30000000

    # tbyte
    FF_TBYT = 0x40000000

    # ASCII ?
    FF_ASCI = 0x50000000

    # Struct ?
    FF_STRU = 0x60000000

    # octaword/xmm word (16 bytes/128 bits)
    FF_OWRD = 0x70000000

    # float
    FF_FLOAT = 0x80000000

    # double
    FF_DOUBLE = 0x90000000

    # packed decimal real
    FF_PACKREAL = 0xA0000000

    # alignment directive
    FF_ALIGN = 0xB0000000

    # 3-byte data (only with support from the processor module)
    FF_3BYTE = 0xC0000000

    # custom data type
    FF_CUSTOM = 0xD0000000

    # ymm word (32 bytes/256 bits)
    FF_YWRD = 0xE0000000

    # bytes
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_f__.html

    # Mask for byte value.
    MS_VAL = 0x000000FF

    # Byte has value?
    FF_IVL = 0x00000100


class AFLAGS:
    # additional flags
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group___a_f_l__.html

    # has line number info
    AFL_LINNUM = 0x00000001

    # user-defined SP value
    AFL_USERSP = 0x00000002

    # name is public (inter-file linkage)
    AFL_PUBNAM = 0x00000004

    # name is weak
    AFL_WEAKNAM = 0x00000008

    # the item is hidden completely
    AFL_HIDDEN = 0x00000010

    # the instruction/data is specified by the user
    AFL_MANUAL = 0x00000020

    # the code/data border is hidden
    AFL_NOBRD = 0x00000040

    # display struct field name at 0 offset when displaying an offset. More...
    AFL_ZSTROFF = 0x00000080

    # the 1st operand is bitwise negated
    AFL_BNOT0 = 0x00000100

    # the 2nd operand is bitwise negated
    AFL_BNOT1 = 0x00000200

    # item from the standard library. More...
    AFL_LIB = 0x00000400

    # has typeinfo? (NSUP_TYPEINFO)
    AFL_TI = 0x00000800

    # has typeinfo for operand 0? (NSUP_OPTYPES)
    AFL_TI0 = 0x00001000

    # has typeinfo for operand 1? (NSUP_OPTYPES+1)
    AFL_TI1 = 0x00002000

    # has local name too (FF_NAME should be set)
    AFL_LNAME = 0x00004000

    # has type comment? (such a comment may be changed by IDA)
    AFL_TILCMT = 0x00008000

    # toggle leading zeroes for the 1st operand
    AFL_LZERO0 = 0x00010000

    # toggle leading zeroes for the 2nd operand
    AFL_LZERO1 = 0x00020000

    # has user defined instruction color?
    AFL_COLORED = 0x00040000

    # terse structure variable display?
    AFL_TERSESTR = 0x00080000

    # code: toggle sign of the 1st operand
    AFL_SIGN0 = 0x00100000

    # code: toggle sign of the 2nd operand
    AFL_SIGN1 = 0x00200000

    # for imported function pointers: doesn't return. More...
    AFL_NORET = 0x00400000

    # sp delta value is fixed by analysis. More...
    AFL_FIXEDSPD = 0x00800000

    # the previous insn was created for alignment purposes only
    AFL_ALIGNFLOW = 0x01000000

    # the type information is definitive. More...
    AFL_USERTI = 0x02000000

    # function returns a floating point value
    AFL_RETFP = 0x04000000

    # insn modifes SP and uses the modified value More...
    AFL_USEMODSP = 0x08000000

    # autoanalysis should not create code here
    AFL_NOTCODE = 0x10000000


class ida_netnode:
    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def netnode(self, *args, **kwargs):
        return idb.netnode.Netnode(self.idb, *args, **kwargs)


class idc:
    def __init__(self, db, api):
        self.idb = db
        self.api = api
        # these will be the capstone disassemblers, lazily loaded.
        # map from bitness (numbers 16, 32, and 64) to capstone disassembler instance
        self.bit_dis = None
        # map from tuple (segment start, end address) to capstone disassembler instance
        self.seg_dis = None

        # apparently this enum changes with bitness.
        # this is annoying.
        # so, be sure to reference these via an `idc` *instance*.
        # yes:
        #
        #    idc(some_idb).FUNCATTR_START
        #
        # no:
        #
        #    idc.FUNCATTR_START
        #
        # via:
        # https://github.com/zachriggle/idapython/blob/37d2fd13b31fec8e6e53fbb9704fa3cd0cbd5b07/python/idc.py#L4149
        if self.idb.wordsize == 4:
            # function start address
            self.FUNCATTR_START = 0
            # function end address
            self.FUNCATTR_END = 4
            # function flags
            self.FUNCATTR_FLAGS = 8
            # function frame id
            self.FUNCATTR_FRAME = 10
            # size of local variables
            self.FUNCATTR_FRSIZE = 14
            # size of saved registers area
            self.FUNCATTR_FRREGS = 18
            # number of bytes purged from the stack
            self.FUNCATTR_ARGSIZE = 20
            # frame pointer delta
            self.FUNCATTR_FPD = 24
            # function color code
            self.FUNCATTR_COLOR = 28
        elif self.idb.wordsize == 8:
            self.FUNCATTR_START   = 0
            self.FUNCATTR_END     = 8
            self.FUNCATTR_FLAGS   = 16
            self.FUNCATTR_FRAME   = 18
            self.FUNCATTR_FRSIZE  = 26
            self.FUNCATTR_FRREGS  = 34
            self.FUNCATTR_ARGSIZE = 36
            self.FUNCATTR_FPD     = 44
            self.FUNCATTR_COLOR   = 52
            self.FUNCATTR_OWNER   = 18
            self.FUNCATTR_REFQTY  = 26
        else:
            raise RuntimeError('unexpected wordsize')

    def ScreenEA(self):
        return self.api.ScreenEA

    def SegStart(self, ea):
        segs = idb.analysis.Segments(self.idb).segments
        for seg in segs.values():
            if seg.startEA <= ea < seg.endEA:
                return seg.startEA

    def SegEnd(self, ea):
        segs = idb.analysis.Segments(self.idb).segments
        for seg in segs.values():
            if seg.startEA <= ea < seg.endEA:
                return seg.endEA

    def FirstSeg(self):
        segs = idb.analysis.Segments(self.idb).segments
        for startEA in sorted(segs.keys()):
            return startEA

    def NextSeg(self, ea):
        segs = idb.analysis.Segments(self.idb).segments.values()
        segs = sorted(segs, key=lambda s: s.startEA)

        for i, seg in enumerate(segs):
            if seg.startEA <= ea < seg.endEA:
                return segs[i + 1].startEA

    def SegName(self, ea):
        segstrings = idb.analysis.SegStrings(self.idb).strings
        segs = idb.analysis.Segments(self.idb).segments
        for seg in segs.values():
            if seg.startEA <= ea < seg.endEA:
                return segstrings[seg.name_index]

    def MinEA(self):
        segs = idb.analysis.Segments(self.idb).segments.values()
        segs = list(sorted(segs, key=lambda s: s.startEA))
        return segs[0].startEA

    def MaxEA(self):
        segs = idb.analysis.Segments(self.idb).segments.values()
        segs = list(sorted(segs, key=lambda s: s.startEA))
        return segs[-1].endEA

    def GetFlags(self, ea):
        return self.idb.id1.get_flags(ea)

    def IdbByte(self, ea):
        flags = self.GetFlags(ea)
        if self.hasValue(flags):
            return flags & FLAGS.MS_VAL
        else:
            raise KeyError(ea)

    def Head(self, ea):
        flags = self.GetFlags(ea)
        while not self.api.ida_bytes.isHead(flags):
            ea -= 1
            # TODO: handle Index/KeyError here when we overrun a segment
            flags = self.GetFlags(ea)
        return ea

    def ItemSize(self, ea):
        oea = ea
        flags = self.GetFlags(ea)
        if not self.api.ida_bytes.isHead(flags):
            raise ValueError('ItemSize must only be called on a head address.')

        ea += 1
        flags = self.GetFlags(ea)
        while not self.api.ida_bytes.isHead(flags):
            ea += 1
            # TODO: handle Index/KeyError here when we overrun a segment
            flags = self.GetFlags(ea)
        return ea - oea

    def NextHead(self, ea):
        ea += 1
        flags = self.GetFlags(ea)
        while not self.api.ida_bytes.isHead(flags):
            ea += 1
            # TODO: handle Index/KeyError here when we overrun a segment
            flags = self.GetFlags(ea)
        return ea

    def PrevHead(self, ea):
        ea = self.Head(ea)
        ea -= 1
        return self.Head(ea)

    def GetManyBytes(self, ea, size, use_dbg=False):
        '''
        Raises:
          IndexError: if the range extends beyond a segment.
          KeyError: if a byte is not defined.
        '''
        if use_dbg:
            raise NotImplementedError()

        if self.SegStart(ea) != self.SegStart(ea + size):
            raise IndexError((ea, ea + size))

        ret = []
        for i in range(ea, ea + size):
            ret.append(self.IdbByte(i))
        if six.PY2:
            return ''.join(map(chr, ret))
        else:
            return bytes(ret)

    def _load_dis(self):
        if self.seg_dis is not None:
            return
        self.bit_dis = {}
        self.seg_dis = {}

        try:
            import capstone
        except:
            raise RuntimeError("capstone is not installed, use the original version of python-idb")

        PROC_CS_MAP = {
            # this is probably the only platform that is thoroughly tested in python-idb.
            # at least, today.
            'metapc':   capstone.CS_ARCH_X86,    # - Disassemble all IBMPC opcodes

            # the rest of these are probably pretty much untested.
            '8086':     capstone.CS_ARCH_X86,    # - Intel 8086
            '80286r':   capstone.CS_ARCH_X86,    # - Intel 80286 real mode
            '80286p':   capstone.CS_ARCH_X86,    # - Intel 80286 protected mode
            '80386r':   capstone.CS_ARCH_X86,    # - Intel 80386 real mode
            '80386p':   capstone.CS_ARCH_X86,    # - Intel 80386 protected mode
            '80486r':   capstone.CS_ARCH_X86,    # - Intel 80486 real mode
            '80486p':   capstone.CS_ARCH_X86,    # - Intel 80486 protected mode
            '80586r':   capstone.CS_ARCH_X86,    # - Intel Pentium & MMX real mode
            '80586p':   capstone.CS_ARCH_X86,    # - Intel Pentium & MMX prot mode
            '80686p':   capstone.CS_ARCH_X86,    # - Intel Pentium Pro & MMX
            'k62':      capstone.CS_ARCH_X86,    # - AMD K6-2 with 3DNow!
            'p2':       capstone.CS_ARCH_X86,    # - Intel Pentium II
            'p3':       capstone.CS_ARCH_X86,    # - Intel Pentium III
            'athlon':   capstone.CS_ARCH_X86,    # - AMD K7
            'p4':       capstone.CS_ARCH_X86,    # - Intel Pentium 4
            '8085':     capstone.CS_ARCH_X86,    # - Intel 8085
            'ppc':      capstone.CS_ARCH_PPC,    # - PowerPC big endian
            'ppcl':     capstone.CS_ARCH_PPC,    # - PowerPC little endian
            'arm':      capstone.CS_ARCH_ARM,    # - ARM little endian
            'armb':     capstone.CS_ARCH_ARM,    # - ARM big endian
            'mipsl':    capstone.CS_ARCH_MIPS,   # - MIPS little endian
            'mipsb':    capstone.CS_ARCH_MIPS,   # - MIPS big endian
            'mipsrl':   capstone.CS_ARCH_MIPS,   # - MIPS & RSP little
            'mipsr':    capstone.CS_ARCH_MIPS,   # - MIPS & RSP big
            'r5900l':   capstone.CS_ARCH_MIPS,   # - MIPS R5900 little
            'r5900r':   capstone.CS_ARCH_MIPS,   # - MIPS R5900 big
            'sparcb':   capstone.CS_ARCH_SPARC,  # - SPARC big endian
            'sparcl':   capstone.CS_ARCH_SPARC,  # - SPARC little endian

            # i don't think any of the rest of these are supported by capstone today
            'z80':       None,  # - Zilog 80
            'z180':      None,  # - Zilog 180
            'z380':      None,  # - Zilog 380
            '64180':     None,  # - Hitachi HD64180
            'gb':        None,  # - Gameboy
            'z8':        None,  # - Zilog 8
            '860xr':     None,  # - Intel 860 XR
            '860xp':     None,  # - Intel 860 XP
            '8051':      None,  # - Intel 8051
            '80196':     None,  # - Intel 80196
            '80196NP':   None,  # - Intel 80196NP, NU
            'm6502':     None,  # - MOS 6502
            'm65c02':    None,  # - MOS 65c02
            'pdp11':     None,  # - DEC PDP/11
            '68000':     None,  # - Motorola MC68000
            '68010':     None,  # - Motorola MC68010
            '68020':     None,  # - Motorola MC68020
            '68030':     None,  # - Motorola MC68030
            '68040':     None,  # - Motorola MC68040
            '68330':     None,  # - Motorola CPU32 (68330)
            '68882':     None,  # - Motorola MC68020 with MC68882
            '68851':     None,  # - Motorola MC68020 with MC68851
            '68020EX':   None,  # - Motorola MC68020 with both
            'colfire':   None,  # - Motorola ColdFire
            '68K':       None,  # - Motorola MC680x0 all opcodes
            '6800':      None,  # - Motorola MC68HC00
            '6801':      None,  # - Motorola MC68HC01
            '6803':      None,  # - Motorola MC68HC03
            '6301':      None,  # - Hitachi HD 6301
            '6303':      None,  # - Hitachi HD 6303
            '6805':      None,  # - Motorola MC68HC05
            '6808':      None,  # - Motorola MC68HC08
            '6809':      None,  # - Motorola MC68HC09
            '6811':      None,  # - Motorola MC68HC11
            '6812':      None,  # - Motorola MC68HC12
            'hcs12':     None,  # - Motorola MC68HCS12
            '6816':      None,  # - Motorola MC68HC16
            'java':      None,  # - java
            'tms320c2':  None,  # - TMS320C2x series
            'tms320c5':  None,  # - TMS320C5x series
            'tms320c6':  None,  # - TMS320C6x series
            'tms320c3':  None,  # - TMS320C3x series
            'tms32054':  None,  # - TMS320C54xx series
            'tms32055':  None,  # - TMS320C55xx series
            'sh3':       None,  # - Renesas SH-3 (little endian)
            'sh3b':      None,  # - Renesas SH-3 (big endian)
            'sh4':       None,  # - Renesas SH-4 (little endian)
            'sh4b':      None,  # - Renesas SH-4 (big endian)
            'sh2a':      None,  # - Renesas SH-2A (big endian)
            'avr':       None,  # - ATMEL AVR                      (ATMEL family)
            'h8300':     None,  # - H8/300x in normal mode
            'h8300a':    None,  # - H8/300x in advanced mode
            'h8s300':    None,  # - H8S in normal mode
            'h8s300a':   None,  # - H8S in advanced mode
            'h8500':     None,  # - H8/500                         (Hitachi H8/500 family)
            'pic12cxx':  None,  # - Microchip PIC 12-bit (12xxx)
            'pic16cxx':  None,  # - Microchip PIC 14-bit (16xxx)
            'pic18cxx':  None,  # - Microchip PIC 16-bit (18xxx)
            'alphab':    None,  # - DEC Alpha big endian
            'alphal':    None,  # - DEC Alpha little endian
            'hppa':      None,  # - HP PA-RISC big endian
            'dsp56k':    None,  # - Motorola DSP 5600x
            'dsp561xx':  None,  # - Motorola DSP 561xx
            'dsp563xx':  None,  # - Motorola DSP 563xx
            'dsp566xx':  None,  # - Motorola DSP 566xx
            'c166':      None,  # - Siemens C166
            'c166v1':    None,  # - Siemens C166 v1 family
            'c166v2':    None,  # - Siemens C166 v2 family
            'st10':      None,  # - SGS-Thomson ST10
            'super10':   None,  # - Super10
            'st20':      None,  # - SGS-Thomson ST20/C1
            'st20c4':    None,  # - SGS-Thomson ST20/C2-C4
            'st7':       None,  # - SGS-Thomson ST7
            'ia64l':     None,  # - Intel Itanium little endian
            'ia64b':     None,  # - Intel Itanium big endian
            'cli':       None,  # - Microsoft.Net platform
            'net':       None,  # - Microsoft.Net platform (alias)
            'i960l':     None,  # - Intel 960 little endian
            'i960b':     None,  # - Intel 960 big endian
            'f2mc16l':   None,  # - Fujitsu F2MC-16L
            'f2mc16lx':  None,  # - Fujitsu F2MC-16LX
            '78k0':      None,  # - NEC 78k/0
            '78k0s':     None,  # - NEC 78k/0s
            'm740':      None,  # - Mitsubishi 8-bit
            'm7700':     None,  # - Mitsubishi 16-bit
            'm7750':     None,  # - Mitsubishi 16-bit
            'm32r':      None,  # - Mitsubishi 32-bit
            'm32rx':     None,  # - Mitsubishi 32-bit extended
            'st9':       None,  # - STMicroelectronics ST9+
            'fr':        None,  # - Fujitsu FR family
            'm7900':     None,  # - Mitsubishi M7900
            'kr1878':    None,  # - Angstrem KR1878
            'ad218x':    None,  # - Analog Devices ADSP
            'oakdsp':    None,  # - Atmel OAK DSP
            'tricore':   None,  # - Infineon Tricore
            'ebc':       None,  # - EFI Bytecode
            'msp430':    None,  # - Texas Instruments MSP430
        }

        procname = self.api.idaapi.get_inf_structure().procname
        cs_arch = PROC_CS_MAP.get(procname)
        if cs_arch is None:
            raise NotImplementedError('disassembly not supported on arch: ' + procname)

        for seg in idb.analysis.Segments(self.idb).segments.values():
            seg_range = (seg.startEA, seg.endEA)
            if seg.bitness == 0:
                if 16 not in self.bit_dis:
                    self.bit_dis[16] = capstone.Cs(cs_arch, capstone.CS_MODE_16)
                    self.bit_dis[16].detail = True
                self.seg_dis[seg_range] = self.bit_dis[16]
            elif seg.bitness == 1:
                if 32 not in self.bit_dis:
                    self.bit_dis[32] = capstone.Cs(cs_arch, capstone.CS_MODE_32)
                    self.bit_dis[32].detail = True
                self.seg_dis[seg_range] = self.bit_dis[32]
            elif seg.bitness == 2:
                if 64 not in self.bit_dis:
                    self.bit_dis[64] = capstone.Cs(cs_arch, capstone.CS_MODE_64)
                    self.bit_dis[64].detail = True
                self.seg_dis[seg_range] = self.bit_dis[64]
            else:
                raise NotImplementedError('unknown bitness: %d' % (seg.bitness))

    def _disassemble(self, ea):
        size = self.ItemSize(ea)
        buf = self.GetManyBytes(ea, size)
        self._load_dis()

        dis = None
        for (seg_start, seg_end), dis in self.seg_dis.items():
            if seg_start <= ea < seg_end:
                break

        if dis is None:
            raise ValueError('failed to find ea in valid segment: ' + hex(ea))

        try:
            op = next(dis.disasm(buf, ea))
        except StopIteration:
            raise RuntimeError('failed to disassemble %s' % (hex(ea)))
        else:
            return op

    def GetMnem(self, ea):
        op = self._disassemble(ea)
        return op.mnemonic

    def GetDisasm(self, ea):
        op = self._disassemble(ea)
        return '%s\t%s' % (op.mnemonic, op.op_str)

    # one instruction or data
    CIC_ITEM = 1
    # function
    CIC_FUNC = 2
    # segment
    CIC_SEGM = 3
    # default color
    DEFCOLOR = 0xFFFFFFFF

    def GetColor(self, ea, what):
        '''
        Args:
          ea (int): effective address of thing.
          what (int): one of:
            - idc.CIC_ITEM
            - idc.CIC_FUNC
            - idc.CIC_SEGM

        Returns:
          int: the color in RGB. possibly idc.DEFCOLOR if not set.
        '''
        if what != idc.CIC_ITEM:
            raise NotImplementedError()

        if not self.api.ida_nalt.is_colored_item(ea):
            return idc.DEFCOLOR

        nn = self.api.ida_netnode.netnode(ea)
        try:
            return nn.altval(tag='A', index=0x14) - 1
        except KeyError:
            return idc.DEFCOLOR

    def GetFunctionFlags(self, ea):
        func = self.api.ida_funcs.get_func(ea)
        return func.flags

    def GetFunctionAttr(self, ea, attr):
        func = self.api.ida_funcs.get_func(ea)

        if attr == self.FUNCATTR_START:
            return func.startEA
        elif attr == self.FUNCATTR_END:
            return func.endEA
        elif attr == self.FUNCATTR_FLAGS:
            return func.flags
        elif attr == self.FUNCATTR_FRAME:
            return func.frame
        elif attr == self.FUNCATTR_FRSIZE:
            return func.frsize
        elif attr == self.FUNCATTR_FRREGS:
            return func.frregs
        elif attr == self.FUNCATTR_ARGSIZE:
            return func.argsize
        elif attr == self.FUNCATTR_FPD:
            return func.fpd
        elif attr == self.FUNCATTR_COLOR:
            return func.color
        else:
            raise ValueError('unknown attr: %x' % (attr))

    def GetFunctionName(self, ea):
        func = self.api.ida_funcs.get_func(ea)
        # ensure this is a function
        if func.startEA != ea:
            raise KeyError(ea)

        # shouldn't be a chunk
        if is_flag_set(func.flags, func.FUNC_TAIL):
            raise KeyError(ea)

        nn = self.api.ida_netnode.netnode(ea)
        try:
            return nn.name()
        except:
            if self.idb.wordsize == 4:
                return 'sub_%04x' % (ea)
            elif self.idb.wordsize == 8:
                return 'sub_%08x' % (ea)
            else:
                raise RuntimeError('unexpected wordsize')

    def LocByName(self, name):
        try:
            key = ("N" + name).encode('utf-8')
            cursor = self.idb.id0.find(key)
            return idb.netnode.as_uint(cursor.value)
        except KeyError:
            return -1

    def GetInputMD5(self):
        return idb.analysis.Root(self.idb).md5

    def Comment(self, ea):
        return self.api.ida_bytes.get_cmt(ea, False)

    def RptCmt(self, ea):
        return self.api.ida_bytes.get_cmt(ea, True)

    def GetCommentEx(self, ea, repeatable):
        return self.api.ida_bytes.get_cmt(ea, repeatable)

    def GetType(self, ea):
        try:
            f = idb.analysis.Function(self.idb, ea)
        except Exception as e:
            logger.warning('failed to fetch function for GetType: %s', e)
            return None

        try:
            name = f.get_name()
            sig = f.get_signature()
        except KeyError:
            return None

        params = []
        for param in sig.parameters:
            params.append('%s %s' % (param.type, param.name))

        return '{rtype:s} {cc:s} {name:s}({params:s})'.format(
            rtype=sig.rtype,
            cc=sig.calling_convention,
            name=name,
            params=', '.join(params),
        )


    @staticmethod
    def hasValue(flags):
        return flags & FLAGS.FF_IVL > 0

    @staticmethod
    def isDefArg0(flags):
        return flags & FLAGS.MS_0TYPE > 0

    @staticmethod
    def isDefArg1(flags):
        return flags & FLAGS.MS_1TYPE > 0

    @staticmethod
    def isOff0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0CUST

    @staticmethod
    def isOff1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1CUST

    @staticmethod
    def isChar0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0CHAR

    @staticmethod
    def isChar1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1CHAR

    @staticmethod
    def isSeg0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0SEG

    @staticmethod
    def isSeg1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1SEG

    @staticmethod
    def isEnum0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0ENUM

    @staticmethod
    def isEnum1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1ENUM

    @staticmethod
    def isStroff0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0STRO

    @staticmethod
    def isStroff1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1STRO

    @staticmethod
    def isStkvar0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0STK

    @staticmethod
    def isStkvar1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1STK

    @staticmethod
    def isFloat0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0FLT

    @staticmethod
    def isFloat1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1FLT

    @staticmethod
    def isCustFmt0(flags):
        return flags & FLAGS.MS_0TYPE == FLAGS.FF_0CUST

    @staticmethod
    def isCustFmt1(flags):
        return flags & FLAGS.MS_1TYPE == FLAGS.FF_1CUST

    @staticmethod
    def isNum0(flags):
        t = flags & FLAGS.MS_0TYPE
        return t == FLAGS.FF_0NUMB or \
            t == FLAGS.FF_0NUMO or \
            t == FLAGS.FF_0NUMD or \
            t == FLAGS.FF_0NUMH

    @staticmethod
    def isNum1(flags):
        t = flags & FLAGS.MS_1TYPE
        return t == FLAGS.FF_1NUMB or \
            t == FLAGS.FF_1NUMO or \
            t == FLAGS.FF_1NUMD or \
            t == FLAGS.FF_1NUMH

    @staticmethod
    def get_optype_flags0(flags):
        return flags & FLAGS.MS_0TYPE

    @staticmethod
    def get_optype_flags1(flags):
        return flags & FLAGS.MS_1TYPE


class ida_bytes:
    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def get_cmt(self, ea, repeatable):
        flags = self.api.idc.GetFlags(ea)
        if not self.has_cmt(flags):
            raise KeyError(ea)

        nn = self.api.ida_netnode.netnode(ea)
        if repeatable:
            return nn.supstr(tag='S', index=1)
        else:
            return nn.supstr(tag='S', index=0)

    @staticmethod
    def isFunc(flags):
        return flags & FLAGS.MS_CODE == FLAGS.FF_FUNC

    @staticmethod
    def isImmd(flags):
        return flags & FLAGS.MS_CODE == FLAGS.FF_IMMD

    @staticmethod
    def isCode(flags):
        return flags & FLAGS.MS_CLS == FLAGS.FF_CODE

    @staticmethod
    def isData(flags):
        return flags & FLAGS.MS_CLS == FLAGS.FF_DATA

    @staticmethod
    def isTail(flags):
        return flags & FLAGS.MS_CLS == FLAGS.FF_TAIL

    @staticmethod
    def isNotTail(flags):
        return not ida_bytes.isTail(flags)

    @staticmethod
    def isUnknown(flags):
        return flags & FLAGS.MS_CLS == FLAGS.FF_UNK

    @staticmethod
    def isHead(flags):
        return ida_bytes.isCode(flags) or ida_bytes.isData(flags)

    @staticmethod
    def isFlow(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_FLOW > 0

    @staticmethod
    def isVar(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_VAR > 0

    @staticmethod
    def hasExtra(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_LINE > 0

    @staticmethod
    def has_cmt(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_COMM > 0

    @staticmethod
    def hasRef(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_REF > 0

    @staticmethod
    def has_name(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_NAME > 0

    @staticmethod
    def has_dummy_name(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_LABL > 0

    @staticmethod
    def has_auto_name(flags):
        # unknown how to compute this
        raise NotImplementedError()

    @staticmethod
    def has_any_name(flags):
        # unknown how to compute this
        raise NotImplementedError()

    @staticmethod
    def has_user_name(flags):
        # unknown how to compute this
        raise NotImplementedError()

    @staticmethod
    def is_invsign(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_SIGN > 0

    @staticmethod
    def is_bnot(flags):
        return flags & FLAGS.MS_COMM & FLAGS.FF_BNOT > 0

    @staticmethod
    def isByte(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_BYTE

    @staticmethod
    def isWord(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_WORD

    @staticmethod
    def isDwrd(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_DWRD

    @staticmethod
    def isQwrd(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_QWRD

    @staticmethod
    def isOwrd(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_OWRD

    @staticmethod
    def isYwrd(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_YWRD

    @staticmethod
    def isTbyt(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_TBYT

    @staticmethod
    def isFloat(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_FLOAT

    @staticmethod
    def isDouble(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_DOUBLE

    @staticmethod
    def isPackReal(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_PACKREAL

    @staticmethod
    def isASCII(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_ASCI

    @staticmethod
    def isStruct(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_STRU

    @staticmethod
    def isAlign(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_ALIGN

    @staticmethod
    def is3byte(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_3BYTE

    @staticmethod
    def isCustom(flags):
        return flags & FLAGS.DT_TYPE == FLAGS.FF_CUSTOM

    def get_bytes(self, ea, count):
        return self.api.idc.GetManyBytes(ea, count)


class ida_nalt:
    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def get_aflags(self, ea):
        nn = self.api.ida_netnode.netnode(ea)
        try:
            return nn.altval(tag='A', index=0x8)
        except KeyError:
            return 0

    def is_hidden_item(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_HIDDEN)

    def is_hidden_border(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_NOBRD)

    def uses_modsp(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_USEMODSP)

    def is_zstroff(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_ZSTROFF)

    def is__bnot0(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_BNOT0)

    def is__bnot1(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_BNOT1)

    def is_libitem(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_LIB)

    def has_ti(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_TI)

    def has_ti0(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_TI0)

    def has_ti1(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_TI1)

    def has_lname(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_LNAME)

    def is_tilcmt(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_TILCMT)

    def is_usersp(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_USERSP)

    def is_lzero0(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_LZERO0)

    def is_lzero1(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_LZERO1)

    def is_colored_item(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_COLORED)

    def is_terse_struc(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_TERSESTR)

    def is__invsign0(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_SIGN0)

    def is__invsign1(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_SIGN1)

    def is_noret(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_NORET)

    def is_fixed_spd(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_FIXEDSPD)

    def is_align_flow(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_ALIGNFLOW)

    def is_userti(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_USERTI)

    def is_retfp(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_RETFP)

    def is_notcode(self, ea):
        return is_flag_set(self.get_aflags(ea), AFLAGS.AFL_NOTCODE)

    def get_import_module_qty(self):
        return max(idb.analysis.Imports(self.idb).lib_names.keys())

    def get_import_module_name(self, mod_index):
        return idb.analysis.Imports(self.idb).lib_names[mod_index]

    def enum_import_names(self, mod_index, py_cb):
        imps = idb.analysis.Imports(self.idb)

        # dereference the node id stored in the A val
        nnref = imps.lib_netnodes[mod_index]
        nn = idb.netnode.Netnode(self.idb, nnref)

        for funcaddr in nn.sups():
            funcname = nn.supstr(funcaddr)
            if not py_cb(funcaddr, funcname, None):
                return

        # TODO: where to fetch ordinal?


class ida_funcs:
    # via: https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_u_n_c__.html
    # Function doesn't return.
    FUNC_NORET = 0x00000001

    # Far function.
    FUNC_FAR = 0x00000002

    # Library function.
    FUNC_LIB = 0x00000004

    # Static function.
    FUNC_STATICDEF = 0x00000008

    # Function uses frame pointer (BP)
    FUNC_FRAME = 0x00000010

    # User has specified far-ness. More...
    FUNC_USERFAR = 0x00000020

    # A hidden function chunk.
    FUNC_HIDDEN = 0x00000040

    # Thunk (jump) function.
    FUNC_THUNK = 0x00000080

    # BP points to the bottom of the stack frame.
    FUNC_BOTTOMBP = 0x00000100

    # Function 'non-return' analysis must be performed. More...
    FUNC_NORET_PENDING = 0x00200

    # SP-analysis has been performed. More...
    FUNC_SP_READY = 0x00000400

    # 'argsize' field has been validated. More...
    FUNC_PURGED_OK = 0x00004000

    # This is a function tail. More...
    FUNC_TAIL = 0x00008000

    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def get_func(self, ea):
        '''
        get the func_t associated with the given address.
        if the address is not the start of a function (or function tail), then searches
         for a function that contains the given address.
        note: the range search is pretty slow, since we parse everything on-demand.
        '''
        nn = self.api.ida_netnode.netnode('$ funcs')
        try:
            v = nn.supval(tag='S', index=ea)
        except KeyError:
            # search for the given effective address in the function regions.
            # according to [1], `get_func` only searches the primary region, and not all chunks?
            #
            # [1]: http://www.openrce.org/reference_library/ida_sdk_lookup/get_func
            for func in idb.analysis.Functions(self.idb).functions.values():
                if not (func.startEA <= ea < func.endEA):
                    continue

                if is_flag_set(func.flags, self.FUNC_TAIL):
                    return self.get_func(func.owner)
                else:
                    return func

            raise KeyError(ea)
        else:
            func = idb.analysis.func_t(v, wordsize=self.idb.wordsize)
            if is_flag_set(func.flags, self.FUNC_TAIL):
                return self.get_func(func.owner)
            else:
                return func


class BasicBlock(object):
    '''
    interface extracted from: https://raw.githubusercontent.com/gabtremblay/idabearclean/master/idaapi.py
    '''

    def __init__(self, flowchart, startEA, endEA):
        self.fc = flowchart
        self.id = startEA
        self.startEA = startEA
        self.endEA = endEA
        # types are declared here:
        #  https://www.hex-rays.com/products/ida/support/sdkdoc/gdl_8hpp.html#afa6fb2b53981d849d63273abbb1624bd
        # not sure if they are stored in the idb. seems like probably not.
        self.type = NotImplementedError()

    def preds(self):
        for pred in self.fc.preds[self.startEA]:
            yield self.fc.bbs[pred]

    def succs(self):
        for succ in self.fc.succs[self.startEA]:
            yield self.fc.bbs[succ]

    def __str__(self):
        return 'BasicBlock(startEA: 0x%x, endEA: 0x%x)' % (self.startEA, self.endEA)


def is_empty(s):
    for c in s:
        return False
    return True


class idaapi:
    # xref flags
    # via:
    # https://www.hex-rays.com/products/ida/support/sdkdoc/group__xref__type.html#ga78aab6d0d6bd9cb4904bbdbb5ac4fa71

    # unknown – for compatibility with old versions.
    # Should not be used anymore.
    fl_U = 0
    # Call Far
    fl_CF = 0x10
    # Call Near
    fl_CN = 0x11
    # Jump Far.
    fl_JF = 0x12
    # Jump Near.
    fl_JN = 0x13
    # User specified (obsolete)
    fl_USobsolete = 0x14
    # Ordinary flow: used to specify execution flow to the next instruction.
    fl_F = 0x15
    # unknown – for compatibility with old versions.
    # Should not be used anymore.
    dr_U = 0
    # Offset
    # The reference uses 'offset' of data rather than its value OR
    # The reference appeared because the "OFFSET" flag of instruction is set.
    # The meaning of this type is IDP dependent.
    dr_O = 1
    # Write access.
    dr_W = 2
    # Read access.
    dr_R = 3
    # Text (for forced operands only) Name of data is used in manual operand.
    dr_T = 4
    # Informational (a derived java class references its base class informationally)
    dr_I = 5

    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def _find_bb_end(self, ea):
        '''
        Args:
          ea (int): address at which a basic block begins. behavior undefined if its not a block start.

        Returns:
          int: the address of the final instruction in the basic block. it may be the same as the start.
        '''
        if not is_empty(idb.analysis.get_crefs_from(self.idb, ea,
                                                    types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F])):
            return ea

        while True:
            last_ea = ea
            ea = self.api.idc.NextHead(ea)

            flags = self.api.idc.GetFlags(ea)
            if self.api.ida_bytes.hasRef(flags):
                return last_ea

            if self.api.ida_bytes.isFunc(flags):
                return last_ea

            if not self.api.ida_bytes.isFlow(flags):
                return last_ea

            if not is_empty(idb.analysis.get_crefs_from(self.idb, ea,
                                                        types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F])):
                return ea

    def _find_bb_start(self, ea):
        '''
        Args:
          ea (int): address at which a basic block ends. behavior undefined if its not a block end.

        Returns:
          int: the address of the first instruction in the basic block. it may be the same as the end.
        '''
        while True:
            flags = self.api.idc.GetFlags(ea)
            if self.api.ida_bytes.hasRef(flags):
                return ea

            if self.api.ida_bytes.isFunc(flags):
                return ea

            last_ea = ea
            ea = self.api.idc.PrevHead(ea)

            if not is_empty(idb.analysis.get_crefs_from(self.idb, ea,
                                                        types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F])):
                return last_ea

            if not self.api.ida_bytes.isFlow(flags):
                return last_ea

    def _get_flow_preds(self, ea):
        # this is basically CodeRefsTo with flow=True.
        # need to fixup the return types, though.

        flags = self.api.idc.GetFlags(ea)
        if self.api.ida_bytes.isFlow(flags):
            # prev instruction fell through to this insn
            yield idb.analysis.Xref(self.api.idc.PrevHead(ea), ea, idaapi.fl_F)

        # get all the flow xrefs to this instruction.
        # a flow xref is like a fallthrough or jump, not like a call.
        for xref in idb.analysis.get_crefs_to(self.idb, ea,
                                              types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F]):
            yield xref

    def _get_flow_succs(self, ea):
        # this is basically CodeRefsFrom with flow=True.
        # need to fixup the return types, though.

        nextea = self.api.idc.NextHead(ea)
        nextflags = self.api.idc.GetFlags(nextea)
        if self.api.ida_bytes.isFlow(nextflags):
            # instruction falls through to next insn
            yield idb.analysis.Xref(ea, nextea, idaapi.fl_F)

        # get all the flow xrefs from this instruction.
        # a flow xref is like a fallthrough or jump, not like a call.
        for xref in idb.analysis.get_crefs_from(self.idb, ea,
                                                types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F]):
            yield xref

    def FlowChart(self, func):
        '''
        Example::

            f = idaapi.FlowChart(idaapi.get_func(here()))
            for block in f:
                if p:
                    print "%x - %x [%d]:" % (block.startEA, block.endEA, block.id)
                for succ_block in block.succs():
                    if p:
                        print "  %x - %x [%d]:" % (succ_block.startEA, succ_block.endEA, succ_block.id)

                for pred_block in block.preds():
                    if p:
                        print "  %x - %x [%d]:" % (pred_block.startEA, pred_block.endEA, pred_block.id)

        via: https://github.com/EiNSTeiN-/idapython/blob/master/examples/ex_gdl_qflow_chart.py
        '''

        # i have no idea how this data is indexed in the idb.
        # is it even indexed?
        # therefore, let's parse the basic blocks ourselves!

        class _FlowChart:
            def __init__(self, db, api, ea):
                self.idb = db
                logger.debug('creating flowchart for %x', ea)

                # set of startEA
                seen = set([])

                # map from startEA to BasicBlock instance
                bbs_by_start = {}
                # map from endEA to BasicBlock instance
                bbs_by_end = {}

                # map from startEA to set of startEA
                preds = collections.defaultdict(lambda: set([]))
                # map from startEA to set of startEA
                succs = collections.defaultdict(lambda: set([]))

                endEA = api.idaapi._find_bb_end(ea)
                logger.debug('found end. %x -> %x', ea, endEA)
                block = BasicBlock(self, ea, endEA)
                bbs_by_start[ea] = block
                bbs_by_end[endEA] = block

                q = [block]

                while q:
                    logger.debug('iteration')
                    logger.debug('queue: [%s]', ', '.join(map(str, q)))

                    block = q[0]
                    q = q[1:]

                    logger.debug('exploring %s', block)

                    if block.startEA in seen:
                        logger.debug('already seen!')
                        continue
                    logger.debug('new!')
                    seen.add(block.startEA)

                    for xref in api.idaapi._get_flow_preds(block.startEA):
                        if xref.src not in bbs_by_end:
                            pred_start = api.idaapi._find_bb_start(xref.src)
                            pred = BasicBlock(self, pred_start, xref.src)
                            bbs_by_start[pred.startEA] = pred
                            bbs_by_end[pred.endEA] = pred
                        else:
                            pred = bbs_by_end[xref.src]

                        logger.debug('pred: %s', pred)

                        preds[block.startEA].add(pred.startEA)
                        succs[pred.startEA].add(block.startEA)
                        q.append(pred)

                    for xref in api.idaapi._get_flow_succs(block.endEA):
                        if xref.dst not in bbs_by_start:
                            succ_end = api.idaapi._find_bb_end(xref.dst)
                            succ = BasicBlock(self, xref.dst, succ_end)
                            bbs_by_start[succ.startEA] = succ
                            bbs_by_end[succ.endEA] = succ
                        else:
                            succ = bbs_by_start[xref.dst]

                        logger.debug('succ: %s', succ)

                        succs[block.startEA].add(succ.startEA)
                        preds[succ.startEA].add(block.startEA)
                        q.append(succ)

                self.preds = preds
                self.succs = succs
                self.bbs = bbs_by_start

            def __iter__(self):
                for bb in self.bbs.values():
                    yield bb

        return _FlowChart(self.idb, self.api, func.startEA)

    def get_next_fixup_ea(self, ea):
        nn = self.api.ida_netnode.netnode('$ fixups')
        # TODO: this is really bad algorithmically. we should cache.
        for index in nn.sups(tag='S'):
            if ea <= index:
                return index
        raise KeyError(ea)

    def contains_fixups(self, ea, size):
        try:
            next_fixup = self.get_next_fixup_ea(ea)
        except KeyError:
            return False
        else:
            if next_fixup < ea + size:
                return True
            else:
                return False

    def getseg(self, ea):
        segs = idb.analysis.Segments(self.idb).segments
        for seg in segs.values():
            if seg.startEA <= ea < seg.endEA:
                return seg

    def get_segm_name(self, ea):
        return self.api.idc.SegName(ea)

    def get_segm_end(self, ea):
        return self.api.idc.SegEnd(ea)

    def get_inf_structure(self):
        return idb.analysis.Root(self.idb).idainfo


class idautils:
    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def GetInputFileMD5(self):
        return self.api.idc.GetInputMD5()

    def Segments(self):
        return sorted(idb.analysis.Segments(self.idb).segments.keys())

    def Functions(self):
        ret = []
        for ea, func in idb.analysis.Functions(self.idb).functions.items():
            # we won't report chunks
            if is_flag_set(func.flags, func.FUNC_TAIL):
                continue
            ret.append(func.startEA)
        return list(sorted(ret))

    def CodeRefsTo(self, ea, flow):
        if flow:
            flags = self.api.idc.GetFlags(ea)
            if self.api.ida_bytes.isFlow(flags):
                # prev instruction fell through to this insn
                yield self.api.idc.PrevHead(ea)

        # get all the code xrefs to this instruction.
        # a code xref is like a fallthrough or jump, not like a call.
        for xref in idb.analysis.get_crefs_to(self.idb, ea,
                                              types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F]):
            yield xref.src

    def CodeRefsFrom(self, ea, flow):
        if flow:
            nextea = self.api.idc.NextHead(ea)
            nextflags = self.api.idc.GetFlags(nextea)
            if self.api.ida_bytes.isFlow(nextflags):
                # instruction falls through to next insn
                yield nextea

        # get all the code xrefs from this instruction.
        # a code xref is like a fallthrough or jump, not like a call.
        for xref in idb.analysis.get_crefs_from(self.idb, ea,
                                                types=[idaapi.fl_JN, idaapi.fl_JF, idaapi.fl_F]):
            yield xref.dst


class ida_entry:
    def __init__(self, db, api):
        self.idb = db
        self.api = api

    def get_entry_qty(self):
        ents = idb.analysis.EntryPoints(self.idb)
        return len(ents.functions) + len(ents.main_entry)

    def get_entry_ordinal(self, index):
        ents = idb.analysis.EntryPoints(self.idb)
        try:
            return ents.ordinals[index + 1]
        except KeyError:
            # once we enumerate all the exports by ordinal,
            # then wrap into the "main entry".
            # not sure that there can be more than one, but we attempt to deal here.
            return sorted(ents.main_entry)[index - len(ents.functions) - 1]

    def get_entry(self, ordinal):
        # for the "main entry", ordinal is actually an address.
        ents = idb.analysis.EntryPoints(self.idb)
        return ents.functions[ordinal]

    def get_entry_name(self, ordinal):
        ents = idb.analysis.EntryPoints(self.idb)
        try:
            return ents.function_names[ordinal]
        except KeyError:
            # for the "main entry", ordinal is actually an address.
            return ents.main_entry_name[ordinal]

    def get_entry_forwarder(self, ordinal):
        ents = idb.analysis.EntryPoints(self.idb)
        return ents.forwarded_symbols.get(ordinal)


class IDAPython:
    def __init__(self, db, ScreenEA=None):
        self.idb = db
        self.ScreenEA = ScreenEA

        self.idc = idc(db, self)
        self.idaapi = idaapi(db, self)
        self.idautils = idautils(db, self)
        self.ida_funcs = ida_funcs(db, self)
        self.ida_bytes = ida_bytes(db, self)
        self.ida_netnode = ida_netnode(db, self)
        self.ida_nalt = ida_nalt(db, self)
        self.ida_entry = ida_entry(db, self)




import struct
import logging
from collections import namedtuple

import six


logger = logging.getLogger(__name__)



def uint32(i):
    '''
    Convert the given signed number into its 32-bit little endian unsigned number value.

    Example::

        assert uint32(-1) == 0xFFFFFFFF


    Example::

        assert uint32(1) == 1
    '''
    return struct.unpack('>I', struct.pack('>i', i))[0]


def uint64(i):
    '''
    Convert the given signed number into its 64-bit little endian unsigned number value.

    Example::

        assert uint64(-1) == 0xFFFFFFFFFFFFFFFF


    Example::

        assert uint64(1) == 1
    '''
    return struct.unpack('>Q', struct.pack('>q', i))[0]


class TAGS:
    '''
    via: https://www.hex-rays.com/products/ida/support/sdkdoc/group__nn__res.html#gaedcc558fe55e19ebc6e304ba7ad8c4d6
    '''
    ALTVAL = 'A'
    SUPVAL = 'S'
    CHARVAL = 'C'  # this is just a guess...
    HASHVAL = 'H'
    VALUE = 'V'
    NAME = 'N'
    LINK = 'L'


def make_key(nodeid, tag=None, index=None, wordsize=4):
    '''

    Example::

        k = make_key('Root Node')


    Example::

        k = make_key(0x401000, 'X')

    Example::

        k = make_key(0x401000, 'X', 0x4010A24)
    '''
    if wordsize == 4:
        wordformat = 'I'
    elif wordsize == 8:
        wordformat = 'Q'
    else:
        raise ValueError('unexpected wordsize')

    if isinstance(nodeid, six.string_types):
        return b'N' + nodeid.encode('utf-8')

    elif isinstance(nodeid, six.integer_types):
        if tag is None:
            raise ValueError('tag required')
        if not isinstance(tag, str):
            raise ValueError('tag must be a string')
        if len(tag) != 1:
            raise ValueError('tag must be a single character string')

        tag = tag.encode('ascii')

        if index is None:
            return b'.' + struct.pack('>' + wordformat + 'c', nodeid, tag)
        elif index < 0:
            return b'.' + struct.pack('>' + wordformat + 'c' + wordformat.lower(), nodeid, tag, index)
        else:
            return b'.' + struct.pack('>' + wordformat + 'c' + wordformat, nodeid, tag, index)
    else:
        raise ValueError('unexpected type of nodeid: ' + str(type(nodeid)))


ComplexKey = namedtuple('ComplexKey', ['nodeid', 'tag', 'index'])

TAG_LENGTH = 1
KEY_HEADER_LENGTH = 1


def parse_key(buf, wordsize=4):
    if six.indexbytes(buf, 0x0) != 0x2E:
        raise ValueError('buf is not a complex key')

    if wordsize == 4:
        wordformat = 'I'
    elif wordsize == 8:
        wordformat = 'Q'
    else:
        raise ValueError('unexpected wordsize')

    nodeid, tag = struct.unpack_from('>' + wordformat + 'c', buf, 1)
    tag = tag.decode('ascii')
    if len(buf) > TAG_LENGTH + wordsize + KEY_HEADER_LENGTH:
        offset = TAG_LENGTH + KEY_HEADER_LENGTH + wordsize
        index = struct.unpack_from('>' + wordformat, buf, offset)[0]
    else:
        index = None

    return ComplexKey(nodeid, tag, index)


def as_uint(buf, wordsize=None):
    if len(buf) == 1:
        return struct.unpack('<B', buf)[0]
    elif len(buf) == 2:
        return struct.unpack('<H', buf)[0]
    elif len(buf) == 4:
        return struct.unpack('<L', buf)[0]
    elif len(buf) == 8:
        return struct.unpack('<Q', buf)[0]
    else:
        return RuntimeError('unexpected buf size')


def as_int(buf, wordsize=None):
    if len(buf) == 1:
        return struct.unpack('<b', buf)[0]
    elif len(buf) == 2:
        return struct.unpack('<h', buf)[0]
    elif len(buf) == 4:
        return struct.unpack('<l', buf)[0]
    elif len(buf) == 8:
        return struct.unpack('<q', buf)[0]
    else:
        return RuntimeError('unexpected buf size')


def as_string(buf, wordsize=None):
    return bytes(buf).rstrip(b'\x00').decode('utf-8').rstrip('\x00')


# try to implement the methods here:
#
#   https://www.hex-rays.com/products/ida/support/sdkdoc/classnetnode.html


Entry = namedtuple('Entry', ['key', 'parsed_key', 'value'])


class Netnode(object):
    def __init__(self, db, nodeid):
        '''
        Args:
          db (idb.IDB): the IDA Pro database.
          nodeid (Union[str, int]): the node id used to identify the netnode.


        Example::

            nn = Netnode("Root Node")
            print(nn.supval(1303))  # --> "6.95"

        Example::

            nn = Netnode(0x401000)
            for xref in nn.alts(tag='X'):
                print(xref)

        TODO: how to address the following keys:
          - $ MAX LINK
          - $ MAX NODE
          - $ MAX DESC

        these are unaddressable via IDA Pro netnodes, too.
        '''
        self.idb = db
        self.wordsize = self.idb.wordsize
        if self.wordsize == 4:
            self.nodebase = 0xFF000000
        elif self.wordsize == 8:
            self.nodebase = 0xFF00000000000000
        else:
            raise RuntimeError('unexpected wordsize')

        if isinstance(nodeid, six.string_types):
            key = make_key(nodeid, wordsize=self.wordsize)
            cursor = self.idb.id0.find(key)
            self.nodeid = as_uint(cursor.value)
            logger.info('resolved string netnode %s to %x', nodeid, self.nodeid)
        elif isinstance(nodeid, six.integer_types):
            self.nodeid = nodeid
        else:
            raise ValueError('unexpected type for nodeid')

    @staticmethod
    def get_nodebase(db):
        if db.wordsize == 4:
            return 0xFF000000
        elif db.wordsize == 8:
            return 0xFF00000000000000

    def name(self):
        '''
        fetch the name associated with the netnode.
        basically supval(tag='N')

        Returns:
          str: the name stored in the netnode.

        Raises:
          KeyError: if the name for the netnode does not exist.
        '''
        key = make_key(self.nodeid, TAGS.NAME, wordsize=self.wordsize)
        cursor = self.idb.id0.find(key)
        return as_string(cursor.value)

    def get_tag_entries(self, tag=TAGS.SUPVAL):
        '''
        generate the entries for the given tag in this netnode.

        this replaces:
          - *1st
          - *nxt
          - *last
          - *prev

        Yields:
          Entry: an entry (with key and value) under the given tag in this netnode.
        '''
        key = make_key(self.nodeid, tag, wordsize=self.wordsize)

        cursor = self.idb.id0.find_prefix(key)
        while bytes(cursor.key).startswith(key):
            parsed_key = parse_key(cursor.key, wordsize=self.idb.wordsize)
            yield Entry(cursor.key, parsed_key, cursor.value)
            try:
                cursor.next()
            except IndexError:
                break

    def get_val(self, index, tag=TAGS.SUPVAL):
        '''
        fetch a sup/alt/hash/etc value from the netnode.
        the nodeid for this netnode must be an integer/effective address.

        Args:
          index (int): the index of the data to fetch.
          tag (str): single character tag.

        Returns:
          bytes: the raw data.
        '''
        key = make_key(self.nodeid, tag, index, wordsize=self.wordsize)
        cursor = self.idb.id0.find(key)
        return bytes(cursor.value)

    def supval(self, index, tag=TAGS.SUPVAL):
        return self.get_val(index, tag)

    def supstr(self, index, tag=TAGS.SUPVAL):
        return as_string(self.supval(index, tag))

    def sups(self, tag=TAGS.SUPVAL):
        '''
        this replaces:
          - sup1st
          - supnxt
          - suplast
          - supprev
        '''
        for entry in self.get_tag_entries(tag=tag):
            yield entry.parsed_key.index

    def supentries(self, tag=TAGS.SUPVAL):
        for entry in self.get_tag_entries(tag=tag):
            yield entry

    def altval(self, index, tag=TAGS.ALTVAL):
        return as_int(self.get_val(index, tag))

    def alts(self, tag=TAGS.ALTVAL):
        '''
        this replaces:
          - alt1st
          - altnxt
          - altlast
          - altprev
        '''
        for entry in self.get_tag_entries(tag=tag):
            yield entry.parsed_key.index

    def altentries(self, tag=TAGS.ALTVAL):
        for entry in self.get_tag_entries(tag=tag):
            # TODO: cast the value?
            yield entry

    def charval(self, index, tag=TAGS.CHARVAL):
        return as_int(self.get_val(index, tag))

    def chars(self, tag=TAGS.ALTVAL):
        '''
        this replaces:
          - char1st
          - charnxt
          - charlast
          - charprev
        '''
        for entry in self.get_tag_entries(tag=tag):
            yield entry.parsed_key.index

    def charentries(self, tag=TAGS.CHARVAL):
        for entry in self.get_tag_entries(tag=tag):
            yield Entry(entry.key, entry.parsed_key, as_int(entry.value))

    def hashval(self, index, tag=TAGS.HASHVAL):
        '''
        TODO: how is this different from a supval?
        '''
        return self.get_val(index, tag)

    def hashes(self, tag=TAGS.HASHVAL):
        '''
        this replaces:
          - hash1st
          - hashnxt
          - hashlast
          - hashprev
        '''
        for entry in self.get_tag_entries(tag=tag):
            yield entry.parsed_key.index

    def hashentries(self, tag=TAGS.HASHVAL):
        for entry in self.get_tag_entries(tag=tag):
            yield entry

    def valobj(self):
        '''
        fetch the default netnode value.
        this is basically supval(tag='V').
        '''
        key = make_key(self.nodeid, TAGS.VALUE, wordsize=self.wordsize)
        cursor = self.idb.id0.find(key)
        return bytes(cursor.value)

    def valstr(self):
        return as_string(self.valobj())

    def value_exists(self):
        try:
            return self.valobj() is not None
        except KeyError:
            return False

    def long_value(self):
        return as_uint(self.valobj())

    def blobsize(self):
        '''
        TODO: how is this arbitrary data stored?
        '''
        raise NotImplementedError()

    def getblob(self):
        raise NotImplementedError()



'''
lots of inspiration from: https://github.com/nlitsme/pyidbutil
'''
import abc
import zlib
import struct
import logging
import functools
from collections import namedtuple

import vstruct
from vstruct.primitives import v_bytes
from vstruct.primitives import v_uint8
from vstruct.primitives import v_uint16
from vstruct.primitives import v_uint32
from vstruct.primitives import v_uint64

import idb
import idb.netnode


logger = logging.getLogger(__name__)


class FileHeader(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        # list of offsets to section headers.
        # order should line up with the SECTIONS definition (see below).
        self.offsets = []
        # list of checksums of sections.
        # order should line up with the SECTIONS definition.
        self.checksums = []

        self.signature = v_bytes(size=0x4)  # IDA1 | IDA2
        self.unk04 = v_uint16()
        self.offset1 = v_uint64()
        self.offset2 = v_uint64()
        self.unk16 = v_uint32()
        self.sig2 = v_uint32()  # | DD CC BB AA |
        self.version = v_uint16()
        self.offset3 = v_uint64()
        self.offset4 = v_uint64()
        self.offset5 = v_uint64()
        self.checksum1 = v_uint32()
        self.checksum2 = v_uint32()
        self.checksum3 = v_uint32()
        self.checksum4 = v_uint32()
        self.checksum5 = v_uint32()
        self.offset6 = v_uint64()
        self.checksum6 = v_uint32()

    def pcb_version(self):
        if self.version != 0x6:
            raise NotImplementedError('unsupported version: %d' % (self.version))

    def pcb_offset6(self):
        self.offsets.append(self.offset1)
        self.offsets.append(self.offset2)
        self.offsets.append(self.offset3)
        self.offsets.append(self.offset4)
        self.offsets.append(self.offset5)
        self.offsets.append(self.offset6)

    def pcb_checksum6(self):
        self.checksums.append(self.checksum1)
        self.checksums.append(self.checksum2)
        self.checksums.append(self.checksum3)
        self.checksums.append(self.checksum4)
        self.checksums.append(self.checksum5)
        self.checksums.append(self.checksum6)

    def validate(self):
        if self.signature not in (b'IDA1', b'IDA2'):
            raise ValueError('bad signature')
        if self.sig2 != 0xAABBCCDD:
            raise ValueError('bad sig2')
        if self.version != 0x6:
            raise ValueError('unsupported version')
        return True


class COMPRESSION_METHOD:
    NONE = 0
    ZLIB = 2


class SectionHeader(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.compression_method = v_uint8()
        self.length = v_uint64()
        self.is_compressed = False

    def pcb_compression_method(self):
        if self.compression_method == COMPRESSION_METHOD.NONE:
            self.is_compressed = False
        else:
            self.is_compressed = True


class Section(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.header = SectionHeader()
        self._contents = v_bytes()
        self.contents = b''

    def vsEmit(self, **kwargs):
        if self.header.is_compressed:
            raise NotImplementedError('Section may not be serialized because it was compressed')

        vstruct.VStruct.vsEmit(self, **kwargs)

    def pcb_header(self):
        self['_contents'].vsSetLength(self.header.length)

    def pcb__contents(self):
        if not self.header.is_compressed:
            self.contents = self._contents
        else:
            self.contents = zlib.decompress(self._contents)
            logger.debug('decompressed parsed section.')

    def validate(self):
        if self.header.length == 0:
            raise ValueError('zero size')
        return True


# sizeof(BranchEntryPointer)
# sizeof(BranchEntry)
# sizeof(LeafEntry)
# sizeof(LeafEntryPointer)
SIZEOF_ENTRY = 0x6


class BranchEntryPointer(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.page = v_uint32()
        self.offset = v_uint16()


class BranchEntry(vstruct.VStruct):
    def __init__(self, page):
        vstruct.VStruct.__init__(self)
        self.page = page
        self.key_length = v_uint16()
        self.key = v_bytes()
        self.value_length = v_uint16()
        self.value = v_bytes()

    def pcb_key_length(self):
        self['key'].vsSetLength(self.key_length)

    def pcb_value_length(self):
        self['value'].vsSetLength(self.value_length)


class LeafEntryPointer(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.common_prefix = v_uint16()
        self.unk02 = v_uint16()
        self.offset = v_uint16()


class LeafEntry(vstruct.VStruct):
    def __init__(self, key, common_prefix):
        vstruct.VStruct.__init__(self)
        self.pkey = key
        self.common_prefix = common_prefix

        self.key_length = v_uint16()
        self._key = v_bytes()
        self.value_length = v_uint16()
        self.value = v_bytes()

        self.key = None

    def pcb_key_length(self):
        self['_key'].vsSetLength(self.key_length)

    def pcb_value_length(self):
        self['value'].vsSetLength(self.value_length)

    def pcb__key(self):
        self.key = self.pkey[:self.common_prefix] + self._key


class Page(vstruct.VStruct):
    '''
    single node in the b-tree.
    has a bunch of key-value entries that may point to other pages.
    binary search these keys and traverse pointers to efficienty query the index.

    branch node::

                                      +-------------+
        +-----------------------------+ ppointer    |  ----> [ node with keys less than entry1.key]
        | entry1.key | entry1.value   |-------------+
        +-----------------------------+ entry1.page |  ----> [ node with entry1.key < X < entry2.key]
        | entry2.key | entry2.value   |-------------+
        +-----------------------------+ entry2.page |  ----> [ node with entry2.key < X < entry3.key]
        | ...        | ...            |-------------+
        +-----------------------------+ ...         |
        | entryN.key | entryN.value   |-------------+
        +-----------------------------+ entryN.key  |  ----> [ node with keys greater than entryN.key]
                                      +-------------+

    leaf node::

        +-----------------------------+
        | entry1.key | entry1.value   |
        +-----------------------------+
        | entry2.key | entry2.value   |
        +-----------------------------+
        | ...        | ...            |
        +-----------------------------+
        | entryN.key | entryN.value   |
        +-----------------------------+

    '''

    def __init__(self, page_size, page_number):
        vstruct.VStruct.__init__(self)
        self.page_number = page_number
        self.ppointer = v_uint32()
        self.entry_count = v_uint16()
        self.contents = v_bytes(page_size)
        # ordered cache of entries, once loaded.
        self._entries = []

    def is_leaf(self):
        '''
        return True if this is a leaf node.

        Returns:
          bool: True if this is a leaf node.
        '''
        return self.ppointer == 0

    def _load_entries(self):
        if not self._entries:
            key = b''
            for i in range(self.entry_count):
                if self.is_leaf():
                    ptr = LeafEntryPointer()
                    ptr.vsParse(self.contents, offset=i * SIZEOF_ENTRY)

                    entry = LeafEntry(key, ptr.common_prefix)
                    entry.vsParse(self.contents, offset=ptr.offset - SIZEOF_ENTRY)
                else:
                    ptr = BranchEntryPointer()
                    ptr.vsParse(self.contents, offset=i * SIZEOF_ENTRY)

                    entry = BranchEntry(int(ptr.page))
                    entry.vsParse(self.contents, offset=ptr.offset - SIZEOF_ENTRY)
                self._entries.append(entry)
                key = entry.key

    def get_entries(self):
        '''
        generate the entries from this page in order.
        each entry is guaranteed to have the following fields:
          - key
          - value

        Yields:
          Union[BranchEntry, LeafEntry]: the b-tree entries from this page.
        '''
        self._load_entries()
        for entry in self._entries:
            yield entry

    def find_index(self, key):
        '''
        find the index of the exact match, or in the case of a branch node,
         the index of the least-greater entry.
        '''
        # implementation note:
        #  suprisingly, using a binary search here does not substantially improve performance.
        #  this is probably the the dominating operations are parsing and allocating entries.
        #  the linear scan below is simpler to read, so we'll use that until it becomes an issue.
        if self.is_leaf():
            for i, entry in enumerate(self.get_entries()):
                if key == bytes(entry.key):
                    return i
        else:
            for i, entry in enumerate(self.get_entries()):
                entry_key = bytes(entry.key)
                if key == entry_key:
                    return i
                elif key < entry_key:
                    # this is the least-greater entry
                    return i
                else:
                    continue
        raise KeyError(key)

    def get_entry(self, entry_number):
        '''
        get the entry at the given index.

        Arguments:
          entry_number (int): the entry index.

        Returns:
          Union[BranchEntry, LeafEntry]: the b-tree entry.

        Raises:
          KeyError: if the entry number is not in the range of entries.
        '''
        self._load_entries()
        if entry_number >= len(self._entries):
            raise KeyError(entry_number)
        return self._entries[entry_number]

    def validate(self):
        last = None
        for entry in self.get_entries():
            if last is None:
                continue

            if last.key >= entry.key:
                raise ValueError('bad page entry sort order')

            last = entry
        return True


class FindStrategy(object):
    '''
    defines the interface for strategies of searching the btree.

    implementors will provide a `.find()` method that operates on a `Cursor` instance.
    the method will update the cursor as it navigates the btree.
    '''
    __meta__ = abc.ABCMeta

    @abc.abstractmethod
    def find(self, cursor, key):
        raise NotImplementedError()


class ExactMatchStrategy(FindStrategy):
    '''
    strategy used to find the entry with exactly the key provided.
    if the exact key is not found, `KeyError` is raised.
    '''

    def _find(self, cursor, page_number, key):
        page = cursor.index.get_page(page_number)
        cursor.path.append(page)

        is_largest = False
        try:
            entry_number = page.find_index(key)
        except KeyError:
            # an entry larger than the given key is not found.
            # but we know we should be searching this node,
            #  so we must have to recurse into the final page pointer.
            is_largest = True
            entry_number = page.entry_count - 1

        entry = page.get_entry(entry_number)

        if bytes(entry.key) == key:
            cursor.entry = entry
            cursor.entry_number = entry_number
            return
        elif page.is_leaf():
            # no matches!
            raise KeyError(key)
        else:
            if is_largest:
                next_page_number = page.get_entry(page.entry_count - 1).page
            elif entry_number == 0:
                next_page_number = page.ppointer
            else:
                next_page_number = page.get_entry(entry_number - 1).page
            self._find(cursor, next_page_number, key)
            return

    def find(self, cursor, key):
        self._find(cursor, cursor.index.root_page, key)


class PrefixMatchStrategy(FindStrategy):
    '''
    strategy used to find the first entry that begins with the given key.
    it may be an exact match, or an exact match does not exist, and the result starts with the given key.
    if no entries start with the given key, `KeyError` is raised.
    '''

    def _find(self, cursor, page_number, key):
        page = cursor.index.get_page(page_number)
        cursor.path.append(page)

        if page.is_leaf():
            for i, entry in enumerate(page.get_entries()):
                entry_key = bytes(entry.key)
                if entry_key.startswith(key):
                    cursor.entry = entry
                    cursor.entry_number = i
                    return
                elif entry_key > key:
                    # as soon as we reach greater entries, we'll never match
                    break

            # pop the final path entry, cause we know its not here
            cursor.path = cursor.path[:-1]
            raise KeyError(key)
        else:  # is branch node
            next_page = page.ppointer
            for i, entry in enumerate(page.get_entries()):
                entry_key = bytes(entry.key)
                if entry_key == key:
                    cursor.entry = entry
                    cursor.entry_number = i
                    return
                elif entry_key.startswith(key):
                    # this is obviously a good match; however,
                    # there may have been an exact match in the sub-page just prior,
                    # so we need to first check that first.
                    try:
                        return self._find(cursor, next_page, key)
                    except KeyError:
                        cursor.entry = entry
                        cursor.entry_number = i
                        return
                elif entry_key > key:
                    # as soon as we reach greater entries, we'll never match.
                    # so we need to check the sub-page just prior.
                    return self._find(cursor, next_page, key)
                else:
                    next_page = entry.page

            # since we haven't found a matching entry, but we know our matches must be somewhere,
            # we need to search the final sub-page, which contains the greatest entries.
            last_entry = page.get_entry(page.entry_count - 1)
            return self._find(cursor, last_entry.page, key)

    def find(self, cursor, key):
        self._find(cursor, cursor.index.root_page, key)


class RoundDownMatchStrategy(FindStrategy):
    '''
    strategy used to find the matching key, or the key just less than the given key.
    it may be an exact match, or an exact match does not exist,
     and the result is less than the given key.
    if no entries are less than the given key, `KeyError` is raised.
    '''

    def _find(self, cursor, page_number, key):
        page = cursor.index.get_page(page_number)
        cursor.path.append(page)

        if page.is_leaf():
            for i, entry in enumerate(page.get_entries()):
                entry_key = bytes(entry.key)

                if entry_key == key:
                    cursor.entry = entry
                    cursor.entry_number = i
                    return
                elif entry_key > key:
                    if i == 0:
                        # need to handle this at the branch node, or
                        #  if this is the only node, bubbles up.
                        raise KeyError(key)
                    else:
                        cursor.entry = page.get_entry(i - 1)
                        cursor.entry_number = i - 1
            entry_number = page.entry_count - 1
            cursor.entry = page.get_entry(entry_number)
            cursor.entry_number = entry_number
        else:  # is branch node
            for i, entry in enumerate(page.get_entries()):
                entry_key = bytes(entry.key)
                if entry_key == key:
                    cursor.entry = entry
                    cursor.entry_number = i
                    return
                elif entry_key > key:
                    if i == 0:
                        # may raise KeyError, and its meant to bubble all the
                        # way up.
                        return self._find(cursor, page.ppointer, key)
                    else:
                        try:
                            entry = page.get_entry(i - 1)
                            return self._find(cursor, entry.page, key)
                        except KeyError:
                            cursor.entry = entry
                            cursor.entry_number = i - 1
                            return
                else:
                    continue

            try:
                entry = page.get_entry(page.entry_count - 1)
                return self._find(cursor, entry.page, key)
            except KeyError:
                cursor.entry = entry
                cursor.entry_number = page.entry_count - 1
                return

    def find(self, cursor, key):
        self._find(cursor, cursor.index.root_page, key)


class MinKeyStrategy(FindStrategy):
    '''
    strategy used to find the minimum key in the index.
    note: this completely ignores the provided key.
    '''

    def _find(self, cursor, page_number):
        page = cursor.index.get_page(page_number)
        cursor.path.append(page)

        if page.is_leaf():
            entry = page.get_entry(0)
            cursor.entry = entry
            cursor.entry_number = 0
        else:  # is branch node
            return self._find(cursor, page.ppointer)

    def find(self, cursor, _):
        self._find(cursor, cursor.index.root_page)


class MaxKeyStrategy(FindStrategy):
    '''
    strategy used to find the maximum key in the index.
    note: this completely ignores the provided key.
    '''

    def _find(self, cursor, page_number):
        page = cursor.index.get_page(page_number)
        cursor.path.append(page)

        if page.is_leaf():
            entry_number = page.entry_count - 1
            entry = page.get_entry(entry_number)
            cursor.entry = entry
            cursor.entry_number = entry_number
        else:  # is branch node
            entry_number = page.entry_count - 1
            entry = page.get_entry(entry_number)
            return self._find(cursor, entry.page)

    def find(self, cursor, _):
        self._find(cursor, cursor.index.root_page)


EXACT_MATCH = ExactMatchStrategy
PREFIX_MATCH = PrefixMatchStrategy
ROUND_DOWN_MATCH = RoundDownMatchStrategy
MIN_KEY = MinKeyStrategy
MAX_KEY = MaxKeyStrategy


class Cursor(object):
    '''
    represents a particular location in the b-tree.
    can be navigated "forward" and "backwards".
    '''

    def __init__(self, index):
        super(Cursor, self).__init__()
        self.index = index

        # ordered list of pages from root to leaf that we traversed to get to
        # this point
        self.path = []

        # populated once found
        self.entry = None

        self.entry_number = None

    def next(self):
        '''
        traverse to the next entry.
        updates this current cursor instance.

        Raises:
          IndexError: if the entry does not exist. the cursor is in an unknown state afterwards.
        '''
        current_page = self.path[-1]
        if current_page.is_leaf():
            if self.entry_number == current_page.entry_count - 1:
                # complex case: have to traverse up and then around.
                # we are at the end of a leaf node. so we need to go to the parent and find the next entry.
                # we may have to go up multiple parents.
                start_key = self.entry.key

                while True:
                    # pop the current node off the path
                    if len(self.path) <= 1:
                        raise IndexError()
                    self.path = self.path[:-1]

                    current_page = self.path[-1]
                    try:
                        entry_number = current_page.find_index(start_key)
                    except KeyError:
                        # not found, becaues its too big for this node.
                        # so we need to go higher.
                        continue
                    else:
                        # found a valid entry, so lets process it.
                        break

                # entry_number now points to the least-greater entry relative to start key.
                # this should be the entry that points to the page from which we just came.
                # we'll want to return the key from this entry.

                self.entry = current_page.get_entry(entry_number)
                self.entry_number = entry_number
                return

            else:  # is inner entry.
                # simple case: simply increment the entry number in the current node.
                next_entry_number = self.entry_number + 1
                next_entry = current_page.get_entry(next_entry_number)

                self.entry = next_entry
                self.entry_number = next_entry_number
                return
        else:  # is branch node.

            # follow the min-edge down to a leaf, and take the min entry.
            next_page = self.index.get_page(self.entry.page)
            while not next_page.is_leaf():
                self.path.append(next_page)
                next_page = self.index.get_page(next_page.ppointer)

            self.path.append(next_page)
            self.entry = next_page.get_entry(0)
            self.entry_number = 0
            return

    def prev(self):
        '''
        traverse to the previous entry.
        updates this current cursor instance.

        Raises:
          IndexError: if the entry does not exist. the cursor is in an unknown state afterwards.
        '''
        current_page = self.path[-1]
        if current_page.is_leaf():
            if self.entry_number == 0:
                # complex case: have to traverse up and then around.
                # we are at the beginning of a leaf node.
                # so we need to go to the parent and find the prev entry.
                # we may have to go up multiple parents.
                start_key = self.entry.key

                while True:
                    # pop the current node off the path
                    if len(self.path) <= 1:
                        raise IndexError()
                    self.path = self.path[:-1]

                    current_page = self.path[-1]
                    try:
                        entry_number = current_page.find_index(start_key)
                    except KeyError:
                        entry_number = current_page.entry_count

                    if entry_number == 0:
                        # not found, becaues its too small for this node.
                        # so we need to go higher.
                        continue
                    else:
                        break

                # entry_number now points to the least-greater entry relative to start key.
                # this should be the entry that points to the page from which we just came.
                # we'll want to return the key from the entry that is just
                # smaller than this one.

                self.entry = current_page.get_entry(entry_number - 1)
                self.entry_number = entry_number - 1
                return

            else:  # is inner entry.
                # simple case: simply decrement the entry number in the current
                # node.
                next_entry_number = self.entry_number - 1
                next_entry = current_page.get_entry(next_entry_number)

                self.entry = next_entry
                self.entry_number = next_entry_number
                return
        else:  # is branch node.

            # follow the max-edge down to a leaf, and take the max entry.
            current_page = self.path[-1]
            if self.entry_number == 0:
                next_page_number = current_page.ppointer
            else:
                next_page_number = current_page.get_entry(self.entry_number - 1).page

            next_page = self.index.get_page(next_page_number)
            while not next_page.is_leaf():
                self.path.append(next_page)
                next_page = self.index.get_page(next_page.get_entry(next_page.entry_count - 1).page)

            self.path.append(next_page)
            self.entry = next_page.get_entry(next_page.entry_count - 1)
            self.entry_number = next_page.entry_count - 1
            return

    @property
    def key(self):
        return self.entry.key

    @property
    def value(self):
        return self.entry.value


class ID0(vstruct.VStruct):
    '''
    a b-tree index.
    keys and values are arbitrary byte strings.

    use `.find()` to identify a matching entry, and use the resulting cursor
     instance to access the value, or traverse to less/greater entries.
    '''

    def __init__(self, buf, wordsize):
        vstruct.VStruct.__init__(self)
        self.buf = idb.memview(buf)
        self.wordsize = wordsize

        self.next_free_offset = v_uint32()
        self.page_size = v_uint16()
        self.root_page = v_uint32()
        self.record_count = v_uint32()
        self.page_count = v_uint32()
        self.unk12 = v_uint8()
        self.signature = v_bytes(size=0x09)

        self._page_cache = {}

    def get_page_buffer(self, page_number):
        if page_number < 1:
            logger.warning('unexpected page number requested: %d', page_number)

        offset = self.page_size * page_number
        return self.buf[offset:offset + self.page_size]

    def get_page(self, page_number):
        page = self._page_cache.get(page_number, None)
        if page is not None:
            return page

        buf = self.get_page_buffer(page_number)
        page = Page(self.page_size, page_number)
        page.vsParse(buf)

        self._page_cache[page_number] = page
        return page

    def find(self, key, strategy=EXACT_MATCH):
        '''
        Args:
          key (bytes): the index key for which to search.
          strategy (Type[MatchStrategy]): the strategy to use to do the search.
            some possible strategies:
              - EXACT_MATCH (default)
              - PREFIX_MATCH

        Returns:
          cursor: the cursor that points to the match.

        Raises:
          KeyError: if the match failes to find a result.
        '''
        c = Cursor(self)
        s = strategy()
        s.find(c, key)
        return c

    def find_prefix(self, key):
        '''
        convenience shortcut for prefix match search.
        '''
        return self.find(key, strategy=PREFIX_MATCH)

    def get_min(self):
        '''
        find the minimum entry in the index.

        Returns:
          cursor: the cursor that points to the match.
        '''
        return self.find(None, strategy=MIN_KEY)

    def get_max(self):
        '''
        find the maximum entry in the index.

        Returns:
          cursor: the cursor that points to the match.
        '''
        return self.find(None, strategy=MAX_KEY)

    def validate(self):
        if self.signature != b'B-tree v2':
            raise ValueError('bad signature')
        return True


class SegmentBounds(vstruct.VStruct):
    '''
    specifies the range of a segment.
    '''

    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)

        self.wordsize = wordsize
        if wordsize == 4:
            self.v_word = v_uint32
        elif wordsize == 8:
            self.v_word = v_uint64
        else:
            raise RuntimeError('unexpected wordsize')

        self.start = self.v_word()
        self.end = self.v_word()


class ID1(vstruct.VStruct):
    '''
    contains flags for each byte.
    '''
    PAGE_SIZE = 0x2000

    def __init__(self, wordsize, buf=None):
        vstruct.VStruct.__init__(self)

        self.wordsize = wordsize
        if wordsize == 4:
            self.v_word = v_uint32
        elif wordsize == 8:
            self.v_word = v_uint64
        else:
            raise RuntimeError('unexpected wordsize')

        self.signature = v_bytes(size=0x04)
        self.unk04 = v_uint32()     # 0x3
        self.segment_count = v_uint32()
        self.unk0C = v_uint32()     # 0x800
        self.page_count = v_uint32()
        # varrays are not actually very list-like,
        #  so the struct field will be ._segments
        #  and the property will be .segments.
        self._segments = vstruct.VArray()
        self.segments = []
        self.padding = v_bytes()
        self.buffer = v_bytes()

    SegmentDescriptor = namedtuple('SegmentDescriptor', ['bounds', 'offset'])

    def pcb_segment_count(self):
        self['_segments'].vsAddElements(self.segment_count,
                                        functools.partial(
                                            SegmentBounds,
                                            self.wordsize))

    def pcb__segments(self):
        offset = 0
        for i in range(self.segment_count):
            segment = self._segments[i]
            segment_byte_count = segment.end - segment.start
            segment_length = 4 * segment_byte_count  # each flag entry is a uint32 on all platforms
            self.segments.append(ID1.SegmentDescriptor(segment, offset))
            offset += segment_length
        offset = 0x14 + (self.segment_count * (2 * self.wordsize))
        padsize = ID1.PAGE_SIZE - offset
        self['padding'].vsSetLength(padsize)

    def pcb_page_count(self):
        self['buffer'].vsSetLength(ID1.PAGE_SIZE * self.page_count)

    def get_segment(self, ea):
        '''
        find the segment that contains the given effective address.

        Returns:
          SegmentDescriptor: segment metadata and location.

        Raises:
          KeyError: if the given address is not in a segment.
        '''
        for segment in self.segments:
            if segment.bounds.start <= ea < segment.bounds.end:
                return segment
        raise KeyError(ea)

    def get_next_segment(self, ea):
        '''
        Fetch the next segment.

        Arguments:
          ea (int): an effective address that should fall within a segment.

        Returns:
          int: the effective address of the start of a segment.

        Raises:
          IndexError: if no more segments are found after the given segment.
          KeyError: if the given effective address does not fall within a segment.
        '''
        for i, segment in enumerate(self.segments):
            if segment.bounds.start <= ea < segment.bounds.end:
                if i == len(self.segments):
                    # this is the last segment, there are no more.
                    raise IndexError(ea)
                else:
                    # there's at least one more, and that's the next one.
                    return self.segments[i + 1]
        raise KeyError(ea)

    def get_flags(self, ea):
        '''
        Fetch the flags for the given effective address.

        > Each byte of the program has 32-bit flags (low 8 bits keep the byte value).
        > These 32 bits are used in GetFlags/SetFlags functions.
        via: https://www.hex-rays.com/products/ida/support/idapython_docs/idc-module.html

        Arguments:
          ea (int): the effective address.

        Returns:
          int: the flags for the given address.

        Raises:
          KeyError: if the given address does not fall within a segment.
        '''
        seg = self.get_segment(ea)
        offset = seg.offset + 4 * (ea - seg.bounds.start)
        return struct.unpack_from('<I', self.buffer, offset)[0]

    def validate(self):
        if self.signature != b'VA*\x00':
            raise ValueError('bad signature')
        if self.unk04 != 0x3:
            raise ValueError('unexpected unk04 value')
        if self.unk0C != 0x800:
            raise ValueError('unexpected unk0C value')
        for segment in self.segments:
            if segment.bounds.start > segment.bounds.end:
                raise ValueError('segment ends before it starts')
        return True


class NAM(vstruct.VStruct):
    '''
    contains pointers to named items.
    '''
    PAGE_SIZE = 0x2000

    def __init__(self, wordsize, buf=None):
        vstruct.VStruct.__init__(self)

        self.wordsize = wordsize
        if wordsize == 4:
            self.v_word = v_uint32
            self.word_fmt = "I"
        elif wordsize == 8:
            self.v_word = v_uint64
            self.word_fmt = "Q"
        else:
            raise RuntimeError('unexpected wordsize')

        self.signature = v_bytes(size=0x04)
        self.unk04 = v_uint32()      # 0x3
        self.non_empty = v_uint32()  # (0x1 non-empty) or (0x0 empty)
        self.unk0C = v_uint32()      # 0x800
        self.page_count = v_uint32()
        self.unk14 = self.v_word()   # 0x0
        # this appears to actually be the number of dwords used by the names.
        # so for an .i64, this is 2x the name count.
        self.name_count = v_uint32()
        self.padding = v_bytes(size=NAM.PAGE_SIZE - (6 * 4 + wordsize))
        self.buffer = v_bytes()

    def pcb_page_count(self):
        self['buffer'].vsSetLength(self.page_count * NAM.PAGE_SIZE)

    def validate(self):
        if self.signature != b'VA*\x00':
            raise ValueError('bad signature')
        if self.unk04 != 0x3:
            raise ValueError('unexpected unk04 value')
        if self.non_empty not in (0x0, 0x1):
            raise ValueError('unexpected non_empty value')
        if self.unk0C != 0x800:
            raise ValueError('unexpected unk0C value')
        if self.unk14 != 0x0:
            raise ValueError('unexpected unk14 value')
        return True

    def names(self):
        count = self.name_count
        if self.wordsize == 8:
            count //= 2

        fmt = "<{count:d}{word_fmt:s}".format(count=count, word_fmt=self.word_fmt)
        size = struct.calcsize(fmt)
        if size > len(self.buffer):
            raise ValueError('buffer too small')
        return struct.unpack(fmt, self.buffer[:size])


class TIL(vstruct.VStruct):
    def __init__(self, buf=None, wordsize=4):
        vstruct.VStruct.__init__(self)
        self.wordsize = wordsize
        self.signature = v_bytes(size=0x06)

    def validate(self):
        if self.signature != b'IDATIL':
            raise ValueError('bad signature')
        return True


SectionDescriptor = namedtuple('SectionDescriptor', ['name', 'cls'])

# section order:
#   - id0
#   - id1
#   - nam
#   - seg
#   - til
#   - id2
#
# via: https://github.com/williballenthin/pyidbutil/blob/master/idblib.py#L262
SECTIONS = [
    SectionDescriptor('id0', ID0),
    SectionDescriptor('id1', ID1),
    SectionDescriptor('nam', NAM),
    SectionDescriptor('seg', None),
    SectionDescriptor('til', TIL),
    SectionDescriptor('id2', None),
]


class IDB(vstruct.VStruct):
    def __init__(self, buf):
        vstruct.VStruct.__init__(self)
        # we use a memoryview since we'll take a bunch of read-only subslices.
        self.buf = idb.memview(buf)

        # list of parsed Section instances or None.
        # the entries should line up with the SECTIONS definition.
        self.sections = []

        # these fields will be parsed from self.buf once the header is parsed.
        # they are *not* linearly parsed during .vsParse().
        self.id0 = None  # type: ID0
        self.id1 = None  # type: ID1
        self.nam = None  # type: NAM
        self.seg = None  # type: NotImplemented
        self.til = None  # type: TIL
        self.id2 = None  # type: NotImplemented

        # these are the only true vstruct fields for this struct.
        self.header = FileHeader()

        # updated once header is parsed.
        self.wordsize = 0
        self.uint = ValueError

    def pcb_header(self):
        if self.header.signature == b'IDA1':
            self.wordsize = 4
            self.uint = idb.netnode.uint32
        elif self.header.signature == b'IDA2':
            self.wordsize = 8
            self.uint = idb.netnode.uint64
        else:
            raise RuntimeError('unexpected file signature: %s' % (self.header.signature))

        # TODO: pass along checksum
        for offset in self.header.offsets:
            if offset == 0:
                self.sections.append(None)
                continue

            sectionbuf = self.buf[offset:]
            section = Section()
            section.vsParse(sectionbuf)
            self.sections.append(section)

        for i, sectiondef in enumerate(SECTIONS):
            if i > len(self.sections):
                logger.debug('missing section: %s', sectiondef.name)
                continue

            section = self.sections[i]
            if not section:
                logger.debug('missing section: %s', sectiondef.name)
                continue

            if not sectiondef.cls:
                logger.warn('section class not implemented: %s', sectiondef.name)
                continue

            s = sectiondef.cls(buf=section.contents, wordsize=self.wordsize)
            s.vsParse(section.contents)
            # vivisect doesn't allow you to assign vstructs to
            #  attributes that are not part of the struct,
            # so we need to override and use the default object behavior.
            object.__setattr__(self, sectiondef.name, s)
            logger.debug('parsed section: %s', sectiondef.name)

    def validate(self):
        self.header.validate()
        self.id0.validate()
        self.id1.validate()
        self.nam.validate()
        self.til.validate()
        return True



import types
import struct
import logging
import binascii
import datetime
import itertools
from collections import namedtuple

import six
import vstruct
from vstruct.primitives import v_str
from vstruct.primitives import v_bytes
from vstruct.primitives import v_uint8
from vstruct.primitives import v_uint16
from vstruct.primitives import v_uint32
from vstruct.primitives import v_uint64

import idb
import idb.netnode


logger = logging.getLogger(__name__)


def is_flag_set(flags, flag):
    return flags & flag == flag


def as_unix_timestamp(buf, wordsize=None):
    '''
    parse unix timestamp bytes into a timestamp.
    '''
    q = struct.unpack_from("<I", buf, 0x0)[0]
    return datetime.datetime.utcfromtimestamp(q)


def as_md5(buf, wordsize=None):
    '''
    parse raw md5 bytes into a hex-formatted string.
    '''
    return binascii.hexlify(buf).decode('ascii')


def cast(buf, V, wordsize=None):
    '''
    apply a vstruct class to a sequence of bytes.

    Args:
        buf (bytes): the bytes to parse.
        V (type[vstruct.VStruct]): the vstruct class.

    Returns:
        V: the parsed instance of V.

    Example::

        s = cast(buf, Stat)
        assert s.gid == 0x1000
    '''
    v = V(wordsize=wordsize)
    v.vsParse(buf)
    return v


def as_cast(V):
    '''
    create a partial function that casts buffers to the given vstruct.

    Args:
        V (type[vstruct.VStruct]): the vstruct class.

    Returns:
        callable[bytes]->V: the function that parses buffers into V instances.

    Example::

        S = as_cast(Stat)
        s = S(buf)
        assert s.gid == 0x1000
    '''
    def inner(buf, wordsize=None):
        return cast(buf, V, wordsize=wordsize)
    setattr(inner, 'V', V.__name__)
    return inner


def unpack_dd(buf, offset=0):
    '''
    unpack up to 32-bits using the IDA-specific data packing format.

    Args:
      buf (bytes): the region to parse.
      offset (int): the offset into the region from which to unpack. default: 0.

    Returns:
      (int, int): the parsed dword, and the number of bytes consumed.

    Raises:
      KeyError: if the bounds of the region are exceeded.
    '''
    if offset != 0:
        # this isn't particularly fast... but its more readable.
        buf = buf[offset:]

    header = six.indexbytes(buf, 0x0)
    if header & 0x80 == 0:
        return header, 1
    elif header & 0xC0 != 0xC0:
        return ((header & 0x7F) << 8) + six.indexbytes(buf, 0x1), 2
    else:
        if header & 0xE0 == 0xE0:
            hi = (six.indexbytes(buf, 0x1) << 8) + six.indexbytes(buf, 0x2)
            low = (six.indexbytes(buf, 0x3) << 8) + six.indexbytes(buf, 0x4)
            size = 5
        else:
            hi = (((header & 0x3F) << 8) + six.indexbytes(buf, 0x1))
            low = (six.indexbytes(buf, 0x2) << 8) + six.indexbytes(buf, 0x3)
            size = 4
        return (hi << 16) + low, size


def unpack_dw(buf, offset=0):
    '''
    unpack word.
    '''
    if offset != 0:
        buf = buf[offset:]

    header = six.indexbytes(buf, 0x0)
    if header & 0x80 == 0:
        return header, 1
    elif header & 0xC0 != 0xC0:
        return ((header << 8) + six.indexbytes(buf, 0x1)) & 0x7FFF, 2
    else:
        return (six.indexbytes(buf, 0x1) << 8) + six.indexbytes(buf, 0x2), 3


def unpack_dq(buf, offset=0):
    '''
    unpack qword.
    '''
    if offset != 0:
        buf = buf[offset:]

    dw1, d1 = unpack_dd(buf)
    dw2, d2 = unpack_dd(buf, offset=d1)
    return (dw2 << 32) + dw1, d1 + d2


def unpack_dds(buf):
    offset = 0
    while offset < len(buf):
        val, size = unpack_dd(buf, offset=offset)
        yield val
        offset += size


def unpack_dqs(buf):
    offset = 0
    while offset < len(buf):
        val, size = unpack_dq(buf, offset=offset)
        yield val
        offset += size


class Unpacker:
    def __init__(self, buf, wordsize, offset=0, should_log=False):
        self.offset = offset
        self.wordsize = wordsize
        self.buf = buf
        self.should_log = should_log

    def _do_unpack(self, unpack_fn):
        v, delta = unpack_fn(self.buf, offset=self.offset)
        if self.should_log:
            logger.debug('%s at %x: %x', unpack_fn.__name__, self.offset, v)
        self.offset += delta
        return v

    def dd(self):
        return self._do_unpack(unpack_dd)

    def dq(self):
        return self._do_unpack(unpack_dq)

    def dw(self):
        return self._do_unpack(unpack_dw)

    def addr(self):
        if self.wordsize == 4:
            return self._do_unpack(unpack_dd)
        elif self.wordsize == 8:
            return self._do_unpack(unpack_dq)
        else:
            raise RuntimeError('unexpected wordsize')


Field = namedtuple('Field', ['name', 'tag', 'index', 'cast', 'minver'])
# namedtuple default args.
# via: https://stackoverflow.com/a/18348004/87207
Field.__new__.__defaults__ = (None,) * len(Field._fields)


class IndexType:
    def __init__(self, name):
        self.name = name

    def str(self):
        return self.name.upper()


ALL = IndexType('all')
ADDRESSES = IndexType('addresses')
NUMBERS = IndexType('numbers')
NODES = IndexType('nodes')

VARIABLE_INDEXES = (ALL, ADDRESSES, NUMBERS, NODES)


class _Analysis(object):
    '''
    this is basically a metaclass for analyzers of IDA Pro netnode namespaces (named nodeid).
    provide set of fields, and parse them from netnodes (nodeid, tag, and optional index)
     when accessed.
    '''

    def __init__(self, db, nodeid, fields):
        self.idb = db
        self.nodeid = nodeid
        self.netnode = idb.netnode.Netnode(db, nodeid)
        self.fields = fields

        idb_version = idb.netnode.Netnode(db, 'Root Node').altval(index=-1)

        # note that order of fields is important:
        #   fields with matching minvers override previously defined fields of the same name
        self._fields_by_name = {f.name: f for f in self.fields
                                if (not f.minver) or
                                   (f.minver and
                                    idb_version >= f.minver)}

    def _is_address(self, index):
        '''
        does the given index fall within a segment?
        '''
        try:
            self.idb.id1.get_segment(index)
            return True
        except KeyError:
            return False

    def _is_node(self, index):
        '''
        does the index look like a raw nodeid?
        '''
        if self.idb.wordsize == 4:
            return index & 0xFF000000 == 0xFF000000
        elif self.idb.wordsize == 8:
            return index & 0xFF00000000000000 == 0xFF00000000000000
        else:
            raise RuntimeError('unexpected wordsize')

    def _is_number(self, index):
        '''
        does the index look like not (address or node)?
        '''
        return (not self._is_address(index)) and (not self._is_node(index))

    def __getattr__(self, key):
        '''
        for the given field name, fetch the value from the appropriate netnode.
        if the field matches multiple indices, then return a mapping from index to value.

        Example::

            assert root.version == 695

        Example::

            assert 0x401000 in entrypoints.ordinals

        Example::

            assert entrypoints.ordinals[0] == 'DllMain'

        Args:
          key (str): the name of the field to fetch.

        Returns:
          any: if a parser was provided, then the parsed data.
            otherwise, the bytes associatd with the field.
            if the field matches multiple indices, then the result is mapping from index to value.

        Raises:
          KeyError: if the field does not exist.
        '''
        if key not in self._fields_by_name:
            return super(Analysis, self).__getattr__(key)

        field = self._fields_by_name[key]
        if field.index in VARIABLE_INDEXES:

            if field.index == ADDRESSES:
                nfilter = self._is_address
            elif field.index == NUMBERS:
                nfilter = self._is_number
            elif field.index == NODES:
                nfilter = self._is_node
            elif field.index == ALL:
                nfilter = lambda x: True
            else:
                raise ValueError('unexpected index')

            # indexes are variable, so map them to the values
            ret = {}
            for sup in self.netnode.supentries(tag=field.tag):
                if not nfilter(sup.parsed_key.index):
                    continue

                if field.cast is None:
                    ret[sup.parsed_key.index] = bytes(sup.value)
                else:
                    v = field.cast(bytes(sup.value), wordsize=self.idb.wordsize)
                    ret[sup.parsed_key.index] = v
            return ret
        else:
            # normal field with an explicit index
            v = self.netnode.supval(field.index, tag=field.tag)
            if field.cast is None:
                return bytes(v)
            else:
                return field.cast(bytes(v),
                                  wordsize=self.idb.wordsize)

    def get_field_tag(self, name):
        '''
        get the tag associated with the given field name.

        Example::

            assert root.get_field_tag('version') == 'A'

        Args:
          key (str): the name of the field to fetch.

        Returns:
          str: a single character string tag.
        '''
        return self._fields_by_name[name].tag

    def get_field_index(self, name):
        '''
        get the index associated with the given field name.
        Example::

            assert root.get_field_index('version') == root.db.uint(-1)

        Args:
          key (str): the name of the field to fetch.

        Returns:
          int or IndexType: the index, if its specified.
            otherwise, this will be an `IndexType` that indicates what indices are expected.
        '''
        return self._fields_by_name[name].index


def Analysis(nodeid, fields):
    '''
    build a partial constructor for _Analysis with the given nodeid and fields.

    Example::

        Root = Analysis('Root Node', [Field(...), ...])
        root = Root(some_idb)
        assert root.version == 695
    '''
    def inner(db):
        return _Analysis(db, nodeid, fields)
    return inner


class IdaInfo(vstruct.VStruct):
    # ref: https://www.hex-rays.com/products/ida/support/sdkdoc/structidainfo.html

    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)

        if wordsize == 4:
            v_word = v_uint32
        elif wordsize == 8:
            v_word = v_uint64
        else:
            raise ValueError('unexpected wordsize')

        """
        v7.0:
        nodeid: ff000002 tag: S index: 0x41b994
        00000000: 69 64 61 00 BC 02 6D 65  74 61 70 63 00 00 00 00  ida...metapc....
        00000010: 00 00 00 00 00 00 A3 00  0B 02 00 00 14 00 00 00  ................
        00000020: 0B 00 00 00 00 00 00 00  F7 FF FF DF 03 00 00 00  ................
        00000030: 00 00 00 00 FF FF FF FF  01 00 00 00 95 16 90 68  ...............h
        00000040: 95 16 90 68 FF FF FF FF  FF FF FF FF 00 10 90 68  ...h...........h
        00000050: 30 E2 9D 68 00 10 90 68  30 E2 9D 68 00 10 90 68  0..h...h0..h...h
        00000060: 00 70 9E 68 10 00 00 00  00 00 00 FF 00 00 10 FF  .p.h............
        00000070: 00 00 00 00 00 02 01 0F  0F 00 40 40 00 00 00 00  ..........@@....
        00000080: 00 00 00 00 00 00 00 00  00 00 02 06 67 BE A3 0E  ............g...
        00000090: 07 00 40 06 00 07 00 18  28 00 50 00 54 03 00 00  ..@.....(.P.T...
        000000A0: 01 00 00 00 01 1B 0A 00  00 00 00 00 61 00 00 00  ............a...
        000000B0: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................
        000000C0: 07 00 00 00 00 01 33 04  01 04 00 02 04 08 08 00  ......3.........
        000000D0: 00 00 00 00 00 00 00 00                           ........

        v6.95:
        00000000: 49 44 41 B7 02 6D 65 74  61 70 63 00 00 23 00 0B  IDA..metapc..#..
        00000010: 00 00 00 00 00 00 00 00  00 00 00 00 00 FF FF FF  ................
        00000020: FF FF FF 95 16 90 68 95  16 90 68 00 10 90 68 30  ......h...h...h0
        00000030: E2 9D 68 00 10 90 68 30  E2 9D 68 00 10 90 68 00  ..h...h0..h...h.
        00000040: 70 9E 68 10 00 00 00 0A  00 00 18 00 01 00 00 02  p.h.............
        00000050: 01 01 00 01 02 01 01 00  00 00 00 00 0F 08 00 09  ................
        00000060: 06 00 01 01 1B 07 61 00  00 00 00 00 00 00 00 00  ......a.........
        00000070: 00 00 00 00 00 00 00 00  00 00 00 01 00 00 00 01  ................
        00000080: 01 01 FF FF FF FF 01 00  00 00 FF FF FF FF 67 BE  ..............g.
        00000090: A3 0E 07 00 40 06 07 00  00 00 00 00 00 00 FD BF  ....@...........
        000000A0: 0F 00 28 00 50 00 40 40  00 00 00 00 00 00 00 00  ..(.P.@@........
        000000B0: 00 00 00 00 00 00 02 01  33 04 01 04 00 02 04 08  ........3.......
        000000C0: 14 00 00 00 08 00 00 00  00 00 00 00 00 00 00 00  ................
        000000D0: 00 00 00 00 00 00 00 00  00 00 00 00 00 01 00 00  ................
        000000E0: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................
        000000F0: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................
        """

        self.tag = v_str(size=0x3)  # 'IDA' below 7.0, 'ida' in 7.0
        self.zero = v_bytes(size=0x0)
        self.version = v_uint16()
        self.procname_size = v_bytes(size=0x0)
        self.procname = v_str(size=0x10)
        # TODO: the exact layout, particularly across versions, of the below is unknown.
        #self.s_genflags = v_uint16()
        #self.lflags = v_uint32()
        #self.database_change_count = v_uint32()
        #self.filetype = v_uint16()
        #self.ostype = v_uint16()
        #self.apptype = v_uint16()
        #self.asmtype = v_uint8()
        #self.specsegs = v_uint8()
        #self.af = v_uint32()
        #self.af2 = v_uint32()
        #self.baseaddr = v_word()
        #self.start_ss = v_uint32()
        #self.start_cs = v_uint32()
        #self.start_ip = v_uint32()
        #self.start_ea = v_uint32()
        #self.start_sp = v_uint32()
        #self.main = v_uint32()
        #self.min_ea = v_uint32()
        #self.max_ea = v_uint32()
        #self.omin_ea = v_uint32()
        #self.omax_ea = v_uint32()
        #self.lowoff = v_uint32()
        #self.highoff = v_uint32()
        #self.maxref = v_word()
        # ... and a bunch of other stuff

    def pcb_tag(self):
        if self.tag == 'IDA':
            # under 7.0
            pass
        elif self.tag == 'ida':
            # 7.0
            self['zero'].vsSetLength(0x1)
        else:
            raise NotImplementedError('raise unknown database tag: ' + self.tag)

    def pcb_version(self):
        # 6.95 database upgraded to v7.0b
        # we have a single byte that describes how long the procname is.
        if self.tag == 'IDA' and self.version == 700:
            self['procname_size'].vsSetLength(0x1)

    def pcb_procname_size(self):
        # 6.95 database upgraded to v7.0b
        # we have a single byte that describes how long the procname is.
        if self.procname_size:
            size = six.indexbytes(self.procname_size, 0x0)
            self['procname'].vsSetLength(size)

    @property
    def procName(self):
        return self.procname


Root = Analysis('Root Node', [
    Field('crc',            'A', -5,       idb.netnode.as_int),
    Field('open_count',     'A', -4,       idb.netnode.as_int),
    Field('created',        'A', -2,       as_unix_timestamp),
    Field('version',        'A', -1,       idb.netnode.as_int),
    Field('md5',            'S', 1302,     as_md5),
    Field('version_string', 'S', 1303,     idb.netnode.as_string),
    Field('idainfo',        'S', 0x41b994, as_cast(IdaInfo)),
    Field('input_file_path','V', None,     idb.netnode.as_string)
])


Loader = Analysis('$ loader name', [
    Field('plugin', 'S', 0, idb.netnode.as_string),
    Field('format', 'S', 1, idb.netnode.as_string),
])


# see `scripts/dump_user.py` for intepretation.
OriginalUser = Analysis('$ original user', [
    Field('data', 'S', 0, bytes),
])


# see `scripts/dump_user.py` for intepretation.
User = Analysis('$ user1', [
    Field('data', 'S', 0, bytes),
])


# '$ entry points' maps from ordinal/address to function name.
#
# supvals:
#   format1
#     index: export ordinal
#     value: function name
#   format2
#     index: EA
#     value: function name
EntryPoints = Analysis('$ entry points', [
    Field('ordinals',  'S', NUMBERS, idb.netnode.as_string),
    Field('addresses', 'S', ADDRESSES, idb.netnode.as_string),
    Field('all',       'S', ALL, idb.netnode.as_string),
])


# this works for v6.95.
# for v7.0b, the data looks something like:
#
#     00000000: FF 68 90 10 00 C0 0D A0  00 90 00 00              .h..........
#
# which looks pack_dd/dq to me.
# TODO: need a way to detect versions and switch analysis implementations.

class FileRegion(vstruct.VStruct):
    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)
        if wordsize == 4:
            v_word = v_uint32
        elif wordsize == 8:
            v_word = v_uint64
        else:
            raise ValueError('unexpected wordsize')

        self.start = v_word()
        self.end = v_word()
        self.rva = v_uint32()


class FileRegionV70:
    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)
        self.start = u.addr()
        self.end = self.start + u.addr()
        self.rva = u.addr()


# '$ fileregions' maps from idb segment start address to details about it.
#
# supvals:
#   format1:
#     index: start effective address
#     value:
#       0x0: start effective address
#       0x4: end effective address
#       0x8: rva start?
FileRegions = Analysis('$ fileregions', [
    Field('regions',  'S', ADDRESSES, as_cast(FileRegion)),
    Field('regions',  'S', ADDRESSES, FileRegionV70, minver=700),
])


class func_t:
    FUNC_TAIL = 0x00008000

    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)

        self.startEA = u.addr()
        self.endEA = self.startEA + u.addr()
        self.flags = u.dw()

        self.frame = None
        self.frsize = None
        self.frregs = None
        self.argsize = None
        self.owner = None
        self.refqty = None

        if not is_flag_set(self.flags, func_t.FUNC_TAIL):
            try:
                self.frame = u.addr()
                self.frsize = u.addr()
                self.frregs = u.dw()
                self.argsize = u.addr()
                # there is some other stuff here, based on IDB version/features
            except IndexError:
                # some of these we don't have, so we'll fall back to the default value of None.
                # eg. owner, refqty only present in some idb versions
                # eg. all of these, if high bit of flags not set.
                pass
        else:
            try:
                self.owner = self.startEA - u.addr()
                self.refqty = u.dd()
            except IndexError:
                # see warning note above
                pass


# '$ funcs' maps from function effective address to details about it.
#
# supvals:
#   format1:
#     index: effective address
#     value: func_t
Functions = Analysis('$ funcs', [
    Field('functions', 'S', ADDRESSES, func_t),
])


class PString(vstruct.VStruct):
    '''
    short pascal string, prefixed with single byte length.
    '''

    def __init__(self, length_is_total=True):
        vstruct.VStruct.__init__(self)
        self.length = v_uint8()
        self.s = v_str()
        self.length_is_total = length_is_total

    def pcb_length(self):
        length = self.length
        if self.length_is_total:
            length = length - 1
        self['s'].vsSetLength(length)


class TypeString(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.header = v_uint8()
        self.length = v_uint8()
        self.s = v_str()

    def pcb_header(self):
        if self.header != 0x3D:
            raise RuntimeError('unexpected type header')

    def pcb_length(self):
        length = self.length
        self['s'].vsSetLength(length - 1)


class StructMember:
    def __init__(self, db, nodeid):
        self.idb = db

        self.nodeid = nodeid
        self.netnode = idb.netnode.Netnode(db, self.nodeid)

    def get_name(self):
        return self.netnode.name().partition('.')[2]

    def get_type(self):
        # nodeid: ff000078 tag: S index: 0x3000
        # 00000000: 3D 0A 48 49 4E 53 54 41  4E 43 45 00              =.HINSTANCE.

        v = self.netnode.supval(tag='S', index=0x3000)
        s = TypeString()
        s.vsParse(v)
        return s.s

    def get_enum_id(self):
        return self.altval(tag='A', index=0xB)

    def get_struct_id(self):
        return self.altval(tag='A', index=0x3)

    def get_member_comment(self):
        return self.supstr(tag='S', index=0x0)

    def get_repeatable_member_comment(self):
        return self.supstr(tag='S', index=0x1)

    # TODO: tag='A', index=0x10
    # TODO: tag='S', index=0x9, "ptrseg"

    def __str__(self):
        try:
            typ = self.get_type()
        except KeyError:
            return 'StructMember(name: %s)' % (self.get_name())
        else:
            return 'StructMember(name: %s, type: %s)' % (self.get_name(), typ)


class STRUCT_FLAGS:
    # via: https://www.hex-rays.com/products/ida/support/sdkdoc/group___s_f__.html

    # is variable size structure (varstruct)? More...
    SF_VAR = 0x00000001

    # is a union? More...
    SF_UNION = 0x00000002

    # has members of type "union"?
    SF_HASUNI = 0x00000004

    # don't include in the chooser list
    SF_NOLIST = 0x00000008

    # the structure comes from type library
    SF_TYPLIB = 0x00000010

    # the structure is collapsed
    SF_HIDDEN = 0x00000020

    # the structure is a function frame
    SF_FRAME = 0x00000040

    # alignment (shift amount: 0..31)
    SF_ALIGN = 0x00000F80

    # ghost copy of a local type
    SF_GHOST = 0x00001000


class Struct:
    '''
    Example::

        struc = Struct(idb, 0xFF000075)
        assert struc.get_name() == 'EXCEPTION_INFO'
        assert len(struc.get_members()) == 5
        assert list(struc.get_members())[0].get_type() == 'DWORD'
    '''
    def __init__(self, db, structid):
        self.idb = db

        # if structid doesn't start with 0xFF0000..., add it.
        nodebase = idb.netnode.Netnode.get_nodebase(db)
        if structid < nodebase:
            structid += nodebase

        self.nodeid = structid
        self.netnode = idb.netnode.Netnode(db, self.nodeid)

    def get_members(self):
        v = self.netnode.supval(tag='M', index=0)
        u = Unpacker(v, wordsize=self.idb.wordsize)
        flags = u.dd()
        count = u.dd()

        if not flags & STRUCT_FLAGS.SF_FRAME:
            raise RuntimeError('unexpected frame header')

        for i in range(count):
            nodeid_offset = u.addr()
            _ = u.addr()
            _ = u.addr()
            _ = u.dd()
            _ = u.dd()

            member_nodeid = self.netnode.nodebase + nodeid_offset
            yield StructMember(self.idb, member_nodeid)


def chunks(l, n):
    '''
    Yield successive n-sized chunks from l.
    via: https://stackoverflow.com/a/312464/87207
    '''
    if isinstance(l, types.GeneratorType):
        while True:
            v = list(itertools.islice(l, n))
            if not v:
                return
            yield v
    else:
        i = 0
        while True:
            try:
                v = l[i:i + n]
                yield v
            except IndexError:
                return
            i += n


def pairs(l):
    return chunks(l, 2)


Chunk = namedtuple('Chunk', ['effective_address', 'length'])
FunctionParameter = namedtuple('FunctionParameter', ['type', 'name'])
FunctionSignature = namedtuple('FunctionSignature', ['calling_convention', 'rtype', 'unk', 'parameters'])
StackChangePoint = namedtuple('StackChangePoint', ['effective_address', 'change'])


class Function:
    '''
    Example::

        func = Function(idb, 0x401000)
        assert func.get_name() == 'DllEntryPoint'
        assert func.get_signature() == '... DllEntryPoint(...)'
    '''

    def __init__(self, db, fva):
        self.idb = db
        self.nodeid = fva
        self.netnode = idb.netnode.Netnode(db, self.nodeid)

    def get_name(self):
        try:
            return self.netnode.name()
        except KeyError:
            return 'sub_%X' % (self.nodeid)

    def get_signature(self):
        typebuf = self.netnode.supval(tag='S', index=0x3000)
        namebuf = self.netnode.supval(tag='S', index=0x3001)

        if six.indexbytes(typebuf, 0x0) != 0xC:
            raise RuntimeError('unexpected signature header')

        if six.indexbytes(typebuf, 0x1) == ord('S'):
            # this is just a guess...
            conv = '__stdcall'
        else:
            raise NotImplementedError()

        rtype = TypeString()
        rtype.vsParse(typebuf, offset=2)

        # this is a guess???
        sp_delta = typebuf[2 + len(rtype)]

        params = []
        typeoffset = 0x2 + len(rtype) + 0x1
        nameoffset = 0x0
        while typeoffset < len(typebuf):
            if six.indexbytes(typebuf, typeoffset) == 0x0:
                break
            typename = TypeString()
            typename.vsParse(typebuf, offset=typeoffset)
            typeoffset += len(typename)

            paramname = PString()
            paramname.vsParse(namebuf, offset=nameoffset)
            nameoffset += len(paramname)

            params.append(FunctionParameter(typename.s, paramname.s))

        return FunctionSignature(conv, rtype.s, sp_delta, params)

    def get_chunks(self):
        v = self.netnode.supval(tag='S', index=0x7000)

        # stored as:
        #
        #   first chunk:
        #     effective addr
        #     length
        #   second chunk:
        #     delta from first.ea + first.length
        #     length
        #   third chunk:
        #     delta from second.ea + second.length
        #     length
        #   ...

        last_ea = 0
        last_length = 0

        if self.idb.wordsize == 4:
            unpacker = unpack_dds
        elif self.idb.wordsize == 8:
            unpacker = unpack_dqs
        else:
            raise RuntimeError('unexpected wordsize')

        for delta, length in pairs(unpacker(v)):
            ea = last_ea + last_length + delta
            yield Chunk(ea, length)
            last_ea = ea
            last_length = length

    # S-0x1000: sp change points
    # S-0x4000: register variables
    # S-0x5000: local labels
    # S-0x7000: function tails

    def get_stack_change_points(self):
        # ref: ida.wll@0x100793d0
        v = self.netnode.supval(tag='S', index=0x1000)
        offset = self.nodeid

        if self.idb.wordsize == 4:
            unpacker = unpack_dds
        elif self.idb.wordsize == 8:
            unpacker = unpack_dqs
        else:
            raise RuntimeError('unexpected wordsize')

        for (delta, change) in pairs(unpacker(v)):
            offset += delta
            if change & 1:
                change = change >> 1
            else:
                change = -(change >> 1)

            yield StackChangePoint(offset, change)


Xref = namedtuple('Xref', ['src', 'dst', 'type'])


def _get_xrefs(db, tag, src=None, dst=None, types=None):
    if not src and not dst:
        raise ValueError('one of src or dst must be provided')

    nn = idb.netnode.Netnode(db, src or dst)
    try:
        for entry in nn.charentries(tag=tag):
            if (types and entry.value in types) or (not types):
                if src:
                    yield Xref(src, entry.parsed_key.index, entry.value)
                else:  # have dst
                    yield Xref(entry.parsed_key.index, dst, entry.value)
    except KeyError:
        return


def get_crefs_to(db, ea, types=None):
    '''
    fetches the code references to the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    '''
    return _get_xrefs(db, dst=ea, tag='X', types=types)


def get_crefs_from(db, ea, types=None):
    '''
    fetches the code references from the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    '''
    return _get_xrefs(db, src=ea, tag='x', types=types)


def get_drefs_to(db, ea, types=None):
    '''
    fetches the data references to the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    '''
    return _get_xrefs(db, dst=ea, tag='D', types=types)


def get_drefs_from(db, ea, types=None):
    '''
    fetches the data references from the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    '''
    return _get_xrefs(db, src=ea, tag='d', types=types)


# under v6.95, this works.
class Fixup(vstruct.VStruct):
    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)
        # sizeof() == 0xB (fixed)
        # possible values: 0x0 - 0xC. top bit has some meaning.
        self.type = v_uint8()
        self.unk01 = v_uint16()  # this might be the segment index + 1?
        if wordsize == 4:
            self.offset = v_uint32()
            self.unk07 = v_uint32()
        elif wordsize == 8:
            self.unk03 = v_uint32()
            self.unk07 = v_uint16()
            self.offset = v_uint64()
        else:
            raise ValueError('unexpected wordsize')

    def pcb_type(self):
        if self.type != 0x04:
            raise NotImplementedError(
                'fixup type %x not yet supported' %
                (self.type))

    def get_fixup_length(self):
        if self.type == 0x4:
            return 0x4
        else:
            raise NotImplementedError(
                'fixup type %x not yet supported' %
                (self.type))


class FixupV70:
    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)

        # tbh, don't really know what these fields are...
        self.type = u.dw()
        self.unk1 = u.dd()
        self.unk2 = u.addr()
        self.offset = u.dd()  # strange this is not an offset

        if self.type != 0x8:
            raise NotImplementedError(
                'fixup type %x not yet supported' %
                (self.type))

    def get_fixup_length(self):
        if self.type == 0x8:
            return 0x4
        else:
            raise NotImplementedError(
                'fixup type %x not yet supported' %
                (self.type))


# '$ fixups' maps from fixup start address to details about it.
Fixups = Analysis('$ fixups', [
    Field('fixups', 'S', ADDRESSES, as_cast(Fixup)),
    Field('fixups', 'S', ADDRESSES, FixupV70, minver=700),
])


def parse_seg_strings(buf, wordsize=None):
    strings = []
    offset = 0x0

    while offset < len(buf):
        if buf[offset] == 0x0:
            break

        string = PString(length_is_total=False)
        string.vsParse(buf, offset=offset)
        offset += len(string)
        strings.append(string.s)

    return strings


SegStrings = Analysis('$ segstrings', [
    Field('strings', 'S', 0, parse_seg_strings),
])


class Seg:
    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)

        self.startEA = u.addr()
        self.endEA = self.startEA + u.addr()
        # index into `$ segstrings` array of strings.
        self.name_index = u.dd()

        # via: https://www.hex-rays.com/products/ida/support/sdkdoc/classsegment__t.html
        # use get/set_segm_class() functions
        self.sclass = u.dd()
        # this field is IDP dependent.
        # TODO: needs non-zero test
        self.orgbase = u.dd()
        # Segment flags
        self.flags = u.dd()
        # Segment alignment codes
        self.align = u.dd()
        # Segment combination codes
        self.comb = u.dd()
        # Segment permissions (0 means no information)
        self.perm = u.dd()
        # Number of bits in the segment addressing.
        # if 0: 16 bits
        # if 1: 32 bits
        # if 2: 64 bits
        # see: https://github.com/fireeye/flare-ida/blob/master/python/flare/jayutils.py#L94
        self.bitness = u.dd()
        # segment type (see Segment types). More...
        self.type = u.dd()
        # segment selector - should be unique.
        self.sel = u.dd()
        # default segment register values.
        self.defsr = NotImplementedError()
        # the segment color
        self.color = (u.dd() - 1) & 0xFFFFFFFF


# '$ segs' maps from segment start address to details about it.
#
# supvals:
#   format1:
#     index: start effective address
#     value: pack_dd data.
#         1: startEA
#         2: size
#         3: name index
#         ...
Segments = Analysis('$ segs', [
    # we use `ALL` for the index type because `_is_address` above does not recognize
    #  addresses not backed by flags/bytes in the IDB.
    # there may be segments for the `.bss`, `extern`, etc sections here, and these
    #  do not have associated flags/bytes.
    # therefore, until we fix `is_address`, being slightly imprecise here works better.
    # note: all indexes in the `$ segs` netnode are addresses, so this assumption works ok.
    Field('segments', 'S', ALL, Seg),
])


Imports = Analysis('$ imports', [
    # index: entry number, value: node id
    Field('lib_netnodes', 'A', NUMBERS, idb.netnode.as_uint),
    # index: entry number, value: dll name
    Field('lib_names', 'S', NUMBERS, idb.netnode.as_string),
])


Import = namedtuple('Import', ['library', 'function_name', 'function_address'])


def enumerate_imports(db):
    '''
    enumerate the functions imported by the module in the given database.

    yields:
      Tuple[str, str, int]: library name, function name, function address
    '''
    imps = Imports(db)
    for index, libname in imps.lib_names.items():
        if index == 0xFFFFFFFF:
            continue

        # dereference the node id stored in the A val
        nnref = imps.lib_netnodes[index]
        nn = idb.netnode.Netnode(db, nnref)

        for funcaddr in nn.sups():
            try:
                funcname = nn.supstr(funcaddr)
                yield Import(libname, funcname, funcaddr)
            except KeyError:
                logger.warning('failed to find import supval: %x', funcaddr)
                continue


EntryPoints = Analysis('$ entry points', [
    # index: ordinal, value: address, terminated by index: uint(-1)
    Field('functions', 'A', NUMBERS, idb.netnode.as_uint),
    # index: address, value: address, should be only one?
    Field('main_entry', 'A', ADDRESSES, idb.netnode.as_uint),
    # index: ordinal, value: ordinal
    Field('ordinals', 'I', NUMBERS, idb.netnode.as_uint),
    # index: ordinal, value: string like (NTDLL!Rtl...)
    Field('forwarded_symbols', 'F', NUMBERS, idb.netnode.as_string),
    # index: ordinal, value: string like (Rtl...)
    Field('function_names', 'S', NUMBERS, idb.netnode.as_string),
    # index: address, value: string like (Rtl...), should be only one?
    Field('main_entry_name', 'S', ADDRESSES, idb.netnode.as_string),
])


EntryPoint = namedtuple('EntryPoint', ['name', 'address', 'ordinal', 'forwarded_symbol'])


def enumerate_entrypoints(db):
    '''
    enumerate the entry point functions in the given database.

    yields:
      Tuple[str, int, int, str]: function name, address, ordinal (optional), and forwarded symbol (optional)
    '''
    ents = EntryPoints(db)

    ordinals = ents.ordinals
    forwarded_symbols = ents.forwarded_symbols
    names = ents.function_names
    names.update(ents.main_entry_name)

    for index, addr in ents.functions.items():
        if index == db.uint(-1):
            break
        yield EntryPoint(names.get(index), addr, ordinals.get(index), forwarded_symbols.get(index))

    for index, addr in ents.main_entry.items():
        yield EntryPoint(names.get(index), addr, ordinals.get(index), forwarded_symbols.get(index))



#!/usr/bin/env python
from setuptools import setup, find_packages

# For Testing:
#
# python3.4 setup.py register -r https://testpypi.python.org/pypi
# python3.4 setup.py bdist_wheel upload -r https://testpypi.python.org/pypi
# python3.4 -m pip install -i https://testpypi.python.org/pypi
#
# For Realz:
#
# python3.4 setup.py register
# python3.4 setup.py bdist_wheel upload
# python3.4 -m pip install

setup(
    name='carbonara-idb',
    version='0.4.0',
    description='Pure Python parser for IDA Pro databases (.idb files). Fork used in Guanciale.',
    author='Willi Ballenthin',
    author_email='willi.ballenthin@gmail.com',
    maintainer="Andrea Fioraldi",
    maintainer_email="andreafioraldi@gmail.com",
    url='https://github.com/Carbonara-Project/Carbonara-IDB',
    license='Apache License 2.0',
    install_requires=[
        'six',
        'hexdump',
        'vivisect-vstruct-wb>=1.0.3',
    ],
    packages=find_packages(exclude=['*.tests', '*.tests.*']),
    entry_points={
        "console_scripts": [
        ]
      },

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
    ],
)
